from __future__ import annotations

__all__ = ["BaseCoreCreator", "setup_core_creator"]

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from objectory import AbstractFactory
from torch.nn import Module
from torch.optim import Optimizer

from gravitorch.datasources.base import BaseDataSource
from gravitorch.lr_schedulers.base import LRSchedulerType
from gravitorch.utils.format import str_target_object

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)


class BaseCoreCreator(ABC, metaclass=AbstractFactory):
    """Defines the base class to create some core engine modules.

    In MLTorch, the core engine modules are:

        - data source
        - model
        - optimizer
        - LR scheduler

    Note it is possible to create these core modules without using
    this class.
    """

    @abstractmethod
    def create(
        self, engine: BaseEngine
    ) -> tuple[BaseDataSource, Module, Optimizer | None, LRSchedulerType | None]:
        r"""Creates the core engine modules.

        This method is responsible to register the event handlers
        associated to the core engine modules.

        Args:
            engine (``BaseEngine``): Specifies the engine.

        Returns:
            tuple with 4 values with the following structure:
                - ``gravitorch.datasources.BaseDataSource``: The
                    initialized data source.
                - ``torch.nn.Module``: The instantiated model.
                - ``torch.optim.Optimizer`` or ``None``: The
                    instantiated optimizer or ``None`` if there
                    is no optimizer (evaluation mode only).
                - ``LRSchedulerType`` or ``None``: The instantiated
                    learning rate (LR) scheduler or ``None`` if
                    there is no learning rate scheduler.
        """


def setup_core_creator(creator: BaseCoreCreator | dict) -> BaseCoreCreator:
    r"""Sets up the core engine modules creator.

    The core engine modules creator is instantiated from its
    configuration by using the ``BaseCoreModulesCreator`` factory
    function.

    Args:
        creator (``BaseCoreCreator`` or dict): Specifies the
            core engine modules creator or its configuration.

    Returns:
        ``BaseCoreCreator``: The instantiated core engine
            modules creator.
    """
    if isinstance(creator, dict):
        logger.info(
            "Initializing a core engine modules creator from its configuration... "
            f"{str_target_object(creator)}"
        )
        creator = BaseCoreCreator.factory(**creator)
    return creator



r"""This package contains the model base class and some implemented models."""

__all__ = [
    "BaseModel",
    "VanillaModel",
]

from gravitorch.models.base import BaseModel
from gravitorch.models.vanilla import VanillaModel



r"""This module implements a data summary for ``torch.Tensor``s of type
long."""

__all__ = ["LongTensorDataSummary", "LongTensorSequenceDataSummary"]

from torch import Tensor

from gravitorch.utils.data_summary.discrete import BaseDiscreteDataSummary
from gravitorch.utils.data_summary.sequence import BaseSequenceDataSummary


class LongTensorDataSummary(BaseDiscreteDataSummary[Tensor]):
    r"""Implements a data summary for ``torch.Tensor``s of type long.

    This data summary assumes that the data are discrete numerical
    values. This data summary computes the following statistics:

        - ``count``: the number of values
        - ``num_unique_values``: the number of unique values
        - ``count_{}``: the number of values per unique value
    """

    def add(self, data: Tensor) -> None:
        r"""Adds new data to the summary.

        Args:
            data (``torch.Tensor`` of type long): Specifies the data
                to add to the summary. This method converts the input
                to a ``torch.Tensor`` of type long if the tensor type
                is different.
        """
        if data.numel() > 0:
            self._counter.update(data.long().flatten().tolist())


class LongTensorSequenceDataSummary(BaseSequenceDataSummary[Tensor]):
    r"""Implements a data summary for ``torch.Tensor``s of type long.

    The input should have a shape ``(sequence_length, *)`` where `*`
    means any number of dimensions. This data summary assumes that
    the data are discrete numerical values. This data summary
    computes the following statistics:

        - ``count``: the number of values
        - ``num_unique_values``: the number of unique values
        - ``count_{}``: the number of values per unique value
    """

    def __init__(self):
        super().__init__(value_summary=LongTensorDataSummary())

    def _get_sequence_length(self, data: Tensor) -> int:
        r"""Gets the sequence length of the data.

        Args:
            data (``torch.Tensor`` of type long and shape
                ``(sequence_length, *)`` where `*` means any number of
                dimensions): Specifies the input sequence.

        Returns:
            int: The sequence length.
        """
        return data.shape[0]



__all__ = [
    "absolute_error",
    "absolute_relative_error",
    "asinh_barron_robust_loss",
    "asinh_mse_loss",
    "barron_robust_loss",
    "basic_loss_reduction",
    "check_basic_loss_reduction",
    "log_cosh_loss",
    "msle_loss",
    "relative_mse_loss",
    "relative_smooth_l1_loss",
    "symlog_mse_loss",
    "symmetric_absolute_relative_error",
    "symmetric_relative_smooth_l1_loss",
]

from gravitorch.nn.functional.barron_loss import (
    asinh_barron_robust_loss,
    barron_robust_loss,
)
from gravitorch.nn.functional.error import (
    absolute_error,
    absolute_relative_error,
    symmetric_absolute_relative_error,
)
from gravitorch.nn.functional.loss_helpers import (
    basic_loss_reduction,
    check_basic_loss_reduction,
)
from gravitorch.nn.functional.robust_loss import (
    asinh_mse_loss,
    log_cosh_loss,
    msle_loss,
    relative_mse_loss,
    relative_smooth_l1_loss,
    symlog_mse_loss,
    symmetric_relative_smooth_l1_loss,
)



r"""This module implements some utility functions to check if some packages are
available."""

__all__ = [
    "check_accelerate",
    "check_fairscale",
    "check_matplotlib",
    "check_pillow",
    "check_tensorboard",
    "check_torchvision",
    "is_accelerate_available",
    "is_fairscale_available",
    "is_matplotlib_available",
    "is_pillow_available",
    "is_psutil_available",
    "is_tensorboard_available",
    "is_torchvision_available",
]

from importlib.util import find_spec


def is_accelerate_available() -> bool:
    r"""Indicates if the ``accelerate`` package is installed or not.

    https://huggingface.co/docs/accelerate/index.html
    """
    return find_spec("accelerate") is not None


def check_accelerate() -> None:
    r"""Checks if the ``accelerate`` package is installed.

    Raises:
        RuntimeError if the ``accelerate`` package is not installed.
    """
    if not is_accelerate_available():
        raise RuntimeError(
            "`accelerate` package is required but not installed. "
            "You can install `accelerate` package with the command:\n\n"
            "pip install accelerate\n"
        )


def is_fairscale_available() -> bool:
    r"""Indicates if the ``fairscale`` package is installed or not."""
    return find_spec("fairscale") is not None


def check_fairscale() -> None:
    r"""Checks if the ``fairscale`` package is installed.

    Raises:
        RuntimeError if the ``fairscale`` package is not installed.
    """
    if not is_fairscale_available():
        raise RuntimeError(
            "`fairscale` package is required but not installed. "
            "You can install `fairscale` package with the command:\n\n"
            "pip install fairscale\n"
        )


def is_matplotlib_available() -> bool:
    r"""Indicates if the ``matplotlib`` package is installed or not."""
    return find_spec("matplotlib") is not None


def check_matplotlib() -> None:
    r"""Checks if the ``matplotlib`` package is installed.

    Raises:
        RuntimeError if the ``matplotlib`` package is not installed.
    """
    if not is_matplotlib_available():
        raise RuntimeError(
            "`matplotlib` package is required but not installed. "
            "You can install `matplotlib` package with the command:\n\n"
            "pip install matplotlib\n"
        )


def is_pillow_available() -> bool:
    r"""Indicates if the ``pillow`` package is installed or not."""
    return find_spec("PIL") is not None


def check_pillow() -> None:
    r"""Checks if the pillow package is installed.

    Raises:
        RuntimeError if the ``pillow`` package is not installed.
    """
    if not is_pillow_available():
        raise RuntimeError(
            "`pillow` package is required but not installed. "
            "You can install `pillow` package with the command:\n\n"
            "pip install pillow\n"
        )


def is_psutil_available() -> bool:
    r"""Indicates if the ``psutil`` package is installed or not."""
    return find_spec("psutil") is not None


def is_tensorboard_available() -> bool:
    r"""Indicates if the ``tensorboard`` package is installed or not."""
    return find_spec("tensorboard") is not None


def check_tensorboard() -> None:
    r"""Checks if the ``tensorboard`` package is installed.

    Raises:
        RuntimeError if the ``tensorboard`` package is not installed.
    """
    if not is_tensorboard_available():
        raise RuntimeError(
            "`tensorboard` package is required but not installed. "
            "You can install `tensorboard` package with the command:\n\n"
            "pip install tensorboard\n"
        )


def is_torchvision_available() -> bool:
    r"""Indicates if the ``torchvision`` package is installed or not.

    https://pytorch.org/vision/stable/index.html
    """
    return find_spec("torchvision") is not None


def check_torchvision() -> None:
    r"""Checks if the ``torchvision`` package is installed.

    Raises:
        RuntimeError if the ``torchvision`` package is not installed.
    """
    if not is_torchvision_available():
        raise RuntimeError(
            "`torchvision` package is required but not installed. "
            "You can install `torchvision` package with the command:\n\n"
            "pip install torchvision\n"
        )



__all__ = ["basic_loss_reduction", "check_basic_loss_reduction"]

from torch import Tensor

VALID_REDUCTIONS = ("none", "mean", "sum")


def basic_loss_reduction(tensor: Tensor, reduction: str) -> Tensor:
    r"""Computes a basic reduction for an input tensor.

    This function is designed to be used with loss functions.

    Args:
        tensor (``torch.Tensor``): Specifies the input tensor to
            reduce.
        reduction (str): Specifies the reduction strategy. The valid
            values are ``'mean'``, ``'none'``,  and ``'sum'``.
            ``'none'``: no reduction will be applied, ``'mean'``: the
            sum will be divided by the number of elements in the
            input, ``'sum'``: the output will be summed.

    Returns:
        ``torch.Tensor``: The reduced tensor. The shape of the tensor
            depends on the reduction strategy.

    Raises:
        ValueError if the reduction is not valid.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.nn.functional import basic_loss_reduction
        >>> tensor = torch.arange(6).view(2, 3)
        >>> basic_loss_reduction(tensor, 'none')
        tensor([[0, 1, 2],
                [3, 4, 5]])
        >>> basic_loss_reduction(tensor, 'sum')
        tensor(15)
        >>> basic_loss_reduction(tensor, 'mean')
        tensor(2.5000)
    """
    if reduction == "mean":
        return tensor.float().mean()
    if reduction == "sum":
        return tensor.sum()
    if reduction == "none":
        return tensor
    raise ValueError(
        f"Incorrect reduction: {reduction}. The possible reductions are {VALID_REDUCTIONS}."
    )


def check_basic_loss_reduction(reduction: str) -> None:
    r"""Checks if the provided reduction ia a valid basic loss reduction.

    The valid reduction values are ``'mean'``, ``'none'``,  and
    ``'sum'``.

    Args:
        reduction (str): Specifies the reduction to check.

    Raises:
        ValueError if the provided reduction is not valid.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.nn.functional import check_basic_loss_reduction
        >>>
    """
    if reduction not in VALID_REDUCTIONS:
        raise ValueError(
            f"Incorrect reduction: {reduction}. The valid reductions are: {VALID_REDUCTIONS}"
        )



__all__ = ["setup_loop_observer"]

import logging
from typing import Union

from gravitorch.utils.format import str_target_object
from gravitorch.utils.loop_observers.base import BaseLoopObserver
from gravitorch.utils.loop_observers.noop import NoOpLoopObserver

logger = logging.getLogger(__name__)


def setup_loop_observer(loop_observer: Union[BaseLoopObserver, dict, None]) -> BaseLoopObserver:
    r"""Sets up a loop observer.

    The loop observer is instantiated from its configuration by
    using the ``BaseLoopObserver`` factory function.

    Args:
        loop_observer (``BaseLoopObserver`` or dict or None):
            Specifies the loop observer or its configuration.
            If ``None``, the ``NoOpLoopObserver`` is instantiated.

    Returns:
        ``BaseLoopObserver``: The loop observer.
    """
    if loop_observer is None:
        loop_observer = NoOpLoopObserver()
    if isinstance(loop_observer, dict):
        logger.info(
            "Initializing a loop observer module from its configuration... "
            f"{str_target_object(loop_observer)}"
        )
        loop_observer = BaseLoopObserver.factory(**loop_observer)
    return loop_observer



__all__ = ["ExU"]

from typing import Union

import torch
from torch import Tensor
from torch.nn import Module, Parameter, ReLU
from torch.nn.init import trunc_normal_, zeros_

from gravitorch.nn.utils import setup_nn_module


class ExU(Module):
    r"""Implementation of the exp-centered (ExU) layer.

    This layer was proposed in the following paper:

        Neural Additive Models: Interpretable Machine Learning with
        Neural Nets.
        Agarwal R., Melnick L., Frosst N., Zhang X., Lengerich B.,
        Caruana R., Hinton G.
        NeurIPS 2021. (https://arxiv.org/pdf/2004.13912.pdf)

    Args:
        input_size (int): Specifies the input size.
        output_size (int): Specifies the output size.
        activation (``torch.nn.Module`` or dict or ``None``):
            Specifies the activation layer or its configuration.
            If ``None``, the ReLU layer is used. Default: ``None``
    """

    def __init__(
        self,
        input_size: int,
        output_size: int,
        activation: Union[Module, dict, None] = None,
    ):
        super().__init__()
        self.weight = Parameter(torch.empty(input_size, output_size))
        self.bias = Parameter(torch.zeros(input_size))
        self.activation = setup_nn_module(activation or ReLU())

        self.reset_parameters()

    @property
    def input_size(self) -> int:
        r"""int: The input size"""
        return self.weight.shape[0]

    @property
    def output_size(self) -> int:
        r"""int: The output size"""
        return self.weight.shape[1]

    def extra_repr(self) -> str:
        return f"input_size={self.input_size}, output_size={self.output_size}"

    def forward(self, tensor: Tensor) -> Tensor:
        r"""Computes the forward of the ExU layer.

        Args:
            tensor (``torch.Tensor`` of type float and shape
                ``(d0, d1, ..., dn, input_size)``): Specifies the
                input tensor.

        Returns:
            ``torch.Tensor`` of type float and shape
                ``(d0, d1, ..., dn, output_size)``: The output
                tensor.
        """
        return self.activation(tensor.sub(self.bias).matmul(self.weight.exp()))

    def reset_parameters(self) -> None:
        r"""Resets the parameters.

        As indicated in page 4 of the paper, the weights are initialed
        using a normal distribution ``N(4.0; 0.5)``. The biases are
        initialized to ``0``
        """
        mean, std = 4.0, 0.5
        trunc_normal_(self.weight, mean=mean, std=std, a=mean - 3 * std, b=mean + 3 * std)
        zeros_(self.bias)



__all__ = ["NoParameterInitializer"]

import logging

from gravitorch.engines.base import BaseEngine
from gravitorch.utils.parameter_initializers.base import BaseDefaultParameterInitializer

logger = logging.getLogger(__name__)


class NoParameterInitializer(BaseDefaultParameterInitializer):
    r"""This is the special class that does not update the model parameters.

    You should use this class if the parameters of the models are
    initialized somewhere else.
    """

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def _initialize(self, engine: BaseEngine) -> None:
        logger.info("The model parameters are not updated")



r"""Base module for metrics."""

__all__ = ["BaseMetric", "EmptyMetricError", "setup_metric"]

import logging
from abc import abstractmethod
from typing import Optional, Union

from objectory import AbstractFactory
from torch.nn import Module

from gravitorch.engines.base import BaseEngine
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


class BaseMetric(Module, metaclass=AbstractFactory):
    r"""Defines the base class for the metric.

    This class is used to register the metric using the metaclass
    factory. Child classes must implement the following methods:
        - ``attach``
        - ``forward``
        - ``reset``
        - ``value``
    """

    @abstractmethod
    def attach(self, engine: BaseEngine) -> None:
        r"""Attaches current metric to provided engine.

        This method can be used to:

            - add event handler to the engine
            - set up history trackers

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """

    @abstractmethod
    def forward(self, *args, **kwargs) -> Optional[dict]:
        r"""Updates the metric given a mini-batch of examples.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """

    @abstractmethod
    def reset(self) -> None:
        r"""Resets the metric."""

    @abstractmethod
    def value(self, engine: Optional[BaseEngine] = None) -> dict:
        r"""Evaluates the metric and log the results given all the examples
        previously seen.

        Args:
            engine (``BaseEngine`` or None): Specifies the engine.

        Returns:
             dict: The results of the metric.
        """


class EmptyMetricError(Exception):
    r"""This error is raised when you try to evaluate an empty metric."""


def setup_metric(metric: Union[BaseMetric, dict]) -> BaseMetric:
    r"""Sets up the metric.

    Args:
        metric (``BaseMetric`` or dict): Specifies the metric or its
            configuration.

    Returns:
        ``BaseMetric``: The instantiated metric.
    """
    if isinstance(metric, dict):
        logger.info(f"Initializing a metric from its configuration... {str_target_object(metric)}")
        metric = BaseMetric.factory(**metric)
    return metric



__all__ = ["BaseLoopObserver"]

from abc import ABC, abstractmethod
from typing import Any

from objectory import AbstractFactory

from gravitorch.engines.base import BaseEngine


class BaseLoopObserver(ABC, metaclass=AbstractFactory):
    r"""Defines the base class to implement a loop observer.

    The loop observer is designed to work with both training and
    evaluation loops.
    """

    @abstractmethod
    def start(self, engine: BaseEngine) -> None:
        r"""Resets the observer state at the start of each training or
        evaluation loop.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """

    @abstractmethod
    def end(self, engine: BaseEngine) -> None:
        r"""Performs an action at the end of each training or evaluation loop.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """

    @abstractmethod
    def update(self, engine: BaseEngine, model_input: Any, model_output: Any) -> None:
        r"""Update the observer.

        Args:
            engine (``BaseEngine``): Specifies the engine.
            model_input: Specifies a batch of model input.
            model_output: Specifies a batch of model output.
        """



__all__ = ["TruncNormalParameterInitializer", "recursive_trunc_normal_"]

import logging
from typing import Union

from torch import nn

from gravitorch.engines.base import BaseEngine
from gravitorch.utils.parameter_initializers import BaseDefaultParameterInitializer

logger = logging.getLogger(__name__)


class TruncNormalParameterInitializer(BaseDefaultParameterInitializer):
    r"""Implements a model parameter initializer with a truncated Normal
    strategy.

    Args:
        mean (int or float, optional): Specifies the mean of the
            Normal distribution. Default: ``0.0``
        std (int or float, optional): Specifies the standard
            deviation of the Normal distribution. Default: ``1.0``
        min_cutoff (int or float, optional): Specifies the minimum
            cutoff value. Default: ``-2.0``
        max_cutoff (int or float, optional): Specifies the maximum
            cutoff value. Default: ``2.0``
        learnable_only (bool, optional): If ``True``, only the
            learnable parameters are initialized, otherwise all the
            parameters are initialized. Default: ``True``
        show_stats (bool, optional): If ``True``, the parameter
            statistics are shown at the end of the initialization.
            Default: ``True``
    """

    def __init__(
        self,
        mean: Union[int, float] = 0.0,
        std: Union[int, float] = 1.0,
        min_cutoff: Union[int, float] = -2.0,
        max_cutoff: Union[int, float] = 2.0,
        learnable_only: bool = True,
        show_stats: bool = True,
    ):
        super().__init__(show_stats=show_stats)
        self._mean = float(mean)
        self._std = float(std)
        self._min_cutoff = float(min_cutoff)
        self._max_cutoff = float(max_cutoff)
        self._learnable_only = bool(learnable_only)

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(mean={self._mean}, std={self._std}, "
            f"min_cutoff={self._min_cutoff}, max_cutoff={self._max_cutoff}, "
            f"learnable_only={self._learnable_only})"
        )

    def _initialize(self, engine: BaseEngine) -> None:
        logger.info(
            f"Initialize model parameters with a truncated Normal strategy (mean={self._mean:,.6f}, "
            f"std={self._std:,.6f}, min_cutoff={self._min_cutoff:,.6f}, max_cutoff={self._max_cutoff:,.6f})"
        )
        recursive_trunc_normal_(
            module=engine.model,
            mean=self._mean,
            std=self._std,
            min_cutoff=self._min_cutoff,
            max_cutoff=self._max_cutoff,
            learnable_only=self._learnable_only,
        )


def recursive_trunc_normal_(
    module: nn.Module,
    mean: float = 0.0,
    std: float = 1.0,
    min_cutoff: float = -2.0,
    max_cutoff: float = 2.0,
    learnable_only: bool = True,
) -> None:
    r"""Initialize the parameters of the module with the truncated Normal
    initialization.

    Args:
        module (``torch.nn.Module``): Specifies the module to
            initialize.
        mean (float, optional): Specifies the mean of the Normal
            distribution. Default: ``0.0``
        std (float, optional): Specifies the standard deviation of the
            Normal distribution. Default: ``1.0``
        min_cutoff (float, optional): Specifies the minimum cutoff
            value. Default: ``-2.0``
        max_cutoff (float, optional): Specifies the maximum cutoff
            value. Default: ``2.0``
        learnable_only (bool, optional): If ``True``, only the
            learnable parameters are initialized, otherwise all the
            parameters are initialized. Default: ``True``

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.parameter_initializers import recursive_trunc_normal_
        >>> from torch import nn
        >>> net = nn.Sequential(nn.Linear(4, 6), nn.ReLU(), nn.BatchNorm1d(6), nn.Linear(6, 1))
        >>> recursive_trunc_normal_(net)
    """
    for params in module.parameters():
        if not learnable_only or learnable_only and params.requires_grad:
            nn.init.trunc_normal_(params.data, mean=mean, std=std, a=min_cutoff, b=max_cutoff)



r"""This module implements a no-operation training loop."""

__all__ = ["NoOpTrainingLoop"]

import logging

from gravitorch.engines.base import BaseEngine
from gravitorch.utils.training_loops.base import BaseTrainingLoop

logger = logging.getLogger(__name__)


class NoOpTrainingLoop(BaseTrainingLoop):
    r"""Implements a no-operation training loop.

    This class can be used to ignore the training loop in an engine.
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def train(self, engine: BaseEngine) -> None:
        r"""Trains the model on the training dataset.

        It is a no-operation method.

         Args:
             engine (``BaseEngine``): Specifies the engine.
        """



r"""This module defines the confusion matrix metric for binary labels."""

__all__ = ["BinaryConfusionMatrix", "CategoricalConfusionMatrix"]

import logging
from collections.abc import Sequence
from typing import Optional, Union

from torch import Tensor

from gravitorch.engines.base import BaseEngine
from gravitorch.models.metrics.base import EmptyMetricError
from gravitorch.models.metrics.base_epoch import BaseEpochMetric
from gravitorch.nn import ToCategoricalLabel
from gravitorch.utils.artifacts import PyTorchArtifact
from gravitorch.utils.exp_trackers import EpochStep
from gravitorch.utils.format import str_scalar
from gravitorch.utils.history import MaxScalarHistory, MinScalarHistory
from gravitorch.utils.meters import BinaryConfusionMatrix as BinaryConfusionMatrixMeter
from gravitorch.utils.meters import (
    MulticlassConfusionMatrix as CategoricalConfusionMatrixMeter,
)
from gravitorch.utils.tensor import str_full_tensor

logger = logging.getLogger(__name__)


class BinaryConfusionMatrix(BaseEpochMetric):
    r"""Implements the confusion matrix metric for binary labels.

    Args:
        mode (str): Specifies the mode (e.g. train or eval).
        name (str, optional): Specifies the name of the metric.
            The name is used to log the metric results.
                Default: ``'bin_conf_mat'``
        betas (sequence, optional): Specifies the betas used to
            compute the f-beta score. Default: ``(1,)``
    """

    def __init__(
        self,
        mode: str,
        name: str = "bin_conf_mat",
        betas: Sequence[Union[int, float]] = (1,),
    ):
        super().__init__(mode=mode, name=name)
        self._confusion_matrix = BinaryConfusionMatrixMeter()
        self._betas = tuple(betas)
        self.reset()

    def extra_repr(self) -> str:
        return f"mode={self._mode}, name={self._name}, betas={self._betas}"

    def attach(self, engine: BaseEngine) -> None:
        r"""Attaches current metric to the provided engine.

        This method can be used to:

            - add event handler to the engine
            - set up history trackers

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        super().attach(engine)
        trackers = [
            MaxScalarHistory(name=f"{self._metric_name}_accuracy"),
            MaxScalarHistory(name=f"{self._metric_name}_balanced_accuracy"),
            MinScalarHistory(name=f"{self._metric_name}_false_negative_rate"),
            MinScalarHistory(name=f"{self._metric_name}_false_positive_rate"),
            MaxScalarHistory(name=f"{self._metric_name}_jaccard_index"),
            MaxScalarHistory(name=f"{self._metric_name}_precision"),
            MaxScalarHistory(name=f"{self._metric_name}_recall"),
            MaxScalarHistory(name=f"{self._metric_name}_true_negative_rate"),
            MaxScalarHistory(name=f"{self._metric_name}_true_positive_rate"),
        ]
        for beta in self._betas:
            trackers.append(MaxScalarHistory(name=f"{self._metric_name}_f{beta}_score"))
        for tracker in trackers:
            engine.add_history(tracker)

    def forward(self, prediction: Tensor, target: Tensor) -> None:
        r"""Updates the mean absolute error metric given a mini-batch of
        examples.

        Args:
            prediction (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` or ``(d0, d1, ..., dn, 1)``
                and type float or long): Specifies the predictions.
            target (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` or ``(d0, d1, ..., dn, 1)``
                and type long or float): Specifies the target tensor.
        """
        self._confusion_matrix.update(prediction.flatten(), target.flatten())

    def reset(self) -> None:
        r"""Resets the metric."""
        self._confusion_matrix.reset()

    def value(self, engine: Optional[BaseEngine] = None) -> dict:
        r"""Evaluates the metric and log the results given all the examples
        previously seen.

        Args:
            engine (``BaseEngine``, optional): Specifies the engine.
                This argument is required to log the results in the
                engine. Default: ``None``.

        Returns:
             dict: The results of the metric
        """
        self._confusion_matrix.all_reduce()
        num_predictions = self._confusion_matrix.num_predictions
        if not num_predictions:
            raise EmptyMetricError(f"{self.__class__.__qualname__} is empty")

        logger.info(f"Confusion matrix\n{str_full_tensor(self._confusion_matrix.matrix)}\n")
        results = self._confusion_matrix.compute_all_metrics(
            self._betas, prefix=f"{self._metric_name}_"
        )
        results[f"{self._metric_name}_num_predictions"] = num_predictions
        for name, value in results.items():
            logger.info(f"{name}: {str_scalar(value)}")

        if engine:
            engine.log_metrics(results, EpochStep(engine.epoch))
            engine.create_artifact(
                PyTorchArtifact(
                    tag=f"metric/{self._metric_name}",
                    data={"confusion_matrix": self._confusion_matrix.matrix, "results": results},
                )
            )
        return results


class CategoricalConfusionMatrix(BaseEpochMetric):
    r"""Implements the confusion matrix metric for categorical labels.

    Args:
        mode (str): Specifies the mode (e.g. train or eval).
        num_classes (int): Specifies the number of classes.
        name (str, optional): Specifies the name of the metric.
            The name is used to log the metric results.
            Default: ``'cat_conf_mat'``
        betas (sequence, optional): Specifies the betas used to
            compute the f-beta score. Default: ``(1,)``
    """

    def __init__(
        self,
        mode: str,
        num_classes: int,
        name: str = "cat_conf_mat",
        betas: Sequence[Union[int, float]] = (1,),
    ):
        super().__init__(mode=mode, name=name)
        self.prediction_transform = ToCategoricalLabel()

        self._confusion_matrix = CategoricalConfusionMatrixMeter.from_num_classes(num_classes)
        self._betas = tuple(betas)
        self.reset()

    def extra_repr(self) -> str:
        return (
            f"mode={self._mode}, name={self._name}, "
            f"num_classes={self._confusion_matrix.num_classes}, betas={self._betas}"
        )

    def attach(self, engine: BaseEngine) -> None:
        r"""Attaches current metric to the provided engine.

        This method can be used to:

            - add event handler to the engine
            - set up history trackers

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        super().attach(engine)
        trackers = [
            MaxScalarHistory(name=f"{self._metric_name}_accuracy"),
            MaxScalarHistory(name=f"{self._metric_name}_balanced_accuracy"),
            MaxScalarHistory(name=f"{self._metric_name}_macro_precision"),
            MaxScalarHistory(name=f"{self._metric_name}_macro_recall"),
            MaxScalarHistory(name=f"{self._metric_name}_micro_precision"),
            MaxScalarHistory(name=f"{self._metric_name}_micro_recall"),
            MaxScalarHistory(name=f"{self._metric_name}_weighted_precision"),
            MaxScalarHistory(name=f"{self._metric_name}_weighted_recall"),
        ]
        for beta in self._betas:
            trackers.append(MaxScalarHistory(name=f"{self._metric_name}_macro_f{beta}_score"))
            trackers.append(MaxScalarHistory(name=f"{self._metric_name}_micro_f{beta}_score"))
            trackers.append(MaxScalarHistory(name=f"{self._metric_name}_weighted_f{beta}_score"))
        for tracker in trackers:
            engine.add_history(tracker)

    def forward(self, prediction: Tensor, target: Tensor) -> None:
        r"""Updates the mean absolute error metric given a mini-batch of
        examples.

        Args:
            prediction (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn, num_clasees)`` and type float):
                Specifies the predictions.
            target (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type long or float):
                Specifies the target tensor. The values have to be in
                ``{0, 1, ..., num_classes-1}``.
        """
        self._confusion_matrix.update(
            self.prediction_transform(prediction).flatten(), target.flatten()
        )

    def reset(self) -> None:
        r"""Resets the metric."""
        self._confusion_matrix.reset()

    def value(self, engine: Optional[BaseEngine] = None) -> dict:
        r"""Evaluates the metric and log the results given all the examples
        previously seen.

        Args:
            engine (``BaseEngine``, optional): Specifies the engine.
                This argument is required to log the results in the
                engine. Default: ``None``.

        Returns:
             dict: The results of the metric
        """
        self._confusion_matrix.all_reduce()
        num_predictions = self._confusion_matrix.num_predictions
        if not num_predictions:
            raise EmptyMetricError(f"{self.__class__.__qualname__} is empty")

        logger.info(f"Confusion matrix\n{str_full_tensor(self._confusion_matrix.matrix)}\n")
        results = self._confusion_matrix.compute_scalar_metrics(
            self._betas, prefix=f"{self._metric_name}_"
        )
        results[f"{self._metric_name}_num_predictions"] = num_predictions
        for name, value in results.items():
            logger.info(f"{name}: {str_scalar(value)}")

        if engine:
            engine.log_metrics(results, EpochStep(engine.epoch))
            engine.create_artifact(
                PyTorchArtifact(
                    tag=f"metric/{self._metric_name}",
                    data={
                        "confusion_matrix": self._confusion_matrix.matrix,
                        "results": results,
                        "per_class_metrics": self._confusion_matrix.compute_per_class_metrics(),
                    },
                )
            )
        return results



__all__ = [
    "BaseModelCreator",
    "DataDistributedParallelModelCreator",
    "VanillaModelCreator",
    "setup_model_creator",
]

from gravitorch.creators.model.base import BaseModelCreator, setup_model_creator
from gravitorch.creators.model.ddp import DataDistributedParallelModelCreator
from gravitorch.creators.model.vanilla import VanillaModelCreator



__all__ = ["Clamp", "ClampLoss"]

from typing import Optional, Union

from torch import Tensor
from torch.nn import Module

from gravitorch.nn.functional import basic_loss_reduction, check_basic_loss_reduction
from gravitorch.nn.utils import setup_nn_module


class Clamp(Module):
    r"""Implements a layer to clamp the tensor values.

    Args:
        min_value (float or ``None``, optional): Specifies the
            minimum value. ``None`` means there is no minimum
            value. Default: ``-1.0``
        max_value (float or ``None``, optional): Specifies the
            maximum value. ``None`` means there is no maximum
            value. Default: ``1.0``
    """

    def __init__(self, min_value: Optional[float] = -1.0, max_value: Optional[float] = 1.0):
        super().__init__()
        self._min_value = min_value
        self._max_value = max_value

    def extra_repr(self) -> str:
        return f"min_value={self._min_value}, max_value={self._max_value}"

    def forward(self, tensor: Tensor) -> Tensor:
        r"""Applies the element-wise ReLU-n function.

        Args:
            tensor (``torch.Tensor`` of shape ``(*)``): Specifies the
                input tensor.

        Returns:
            ``torch.Tensor`` with same shape as the input: The output
                tensor.
        """
        return tensor.clamp(min=self._min_value, max=self._max_value)


class ClampLoss(Module):
    r"""Implements a loss function wrapper that clamps output values.

    Args:
        criterion (``torch.nn.Module`` or dict): Specifies the loss
            function to wrap or its configuration.
        min_value (float or ``None``, optional): Specifies the
            minimum value. ``None`` means there is no minimum
            value. Default: ``-1.0``
        max_value (float or ``None``, optional): Specifies the
            maximum value. ``None`` means there is no maximum
            value. Default: ``1.0``
        reduction (string, optional): Specifies the reduction to apply
            to the output: ``'none'`` | ``'mean'`` | ``'sum'``.
            ``'none'``: no reduction will be applied, ``'mean'``: the
            sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be
            summed. Default: ``'mean'``
    """

    def __init__(
        self,
        criterion: Union[Module, dict],
        min_value: Optional[float],
        max_value: Optional[float],
        reduction: str = "none",
    ):
        super().__init__()
        self.criterion = setup_nn_module(criterion)
        self.clamp = Clamp(min_value=min_value, max_value=max_value)

        check_basic_loss_reduction(reduction)
        self.reduction = reduction

    def forward(self, prediction: Tensor, target: Tensor) -> Tensor:
        r"""Computes the loss values, then clamps the values and reduces them.

        Args:
            prediction (``torch.Tensor``): Specifies the predictions.
            target (``torch.Tensor``): Specifies the targets.

        Returns:
            ``torch.Tensor``: The computed loss value.
        """
        return basic_loss_reduction(self.clamp(self.criterion(prediction, target)), self.reduction)



__all__ = ["PickleArtifact"]

from pathlib import Path
from pickle import HIGHEST_PROTOCOL
from typing import Any

from gravitorch.utils.artifacts.base import BaseArtifact
from gravitorch.utils.io import save_pickle


class PickleArtifact(BaseArtifact):
    r"""Implements a Pickle artifact.

    Args:
        tag (str): Specifies the artifact tag. The tag is used to
            define the Pickle filename.
        data: Specifies the data to save in the Pickle artifact. The
            data should be Pickle compatible.
        protocol (int, optional): Specifies the pickle protocol.
            Default: highest protocol available.
    """

    def __init__(self, tag: str, data: Any, protocol: int = HIGHEST_PROTOCOL):
        self._tag = str(tag)
        self._data = data
        self._protocol = protocol

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}(tag={self._tag}, protocol={self._protocol})"

    def create(self, path: Path) -> None:
        save_pickle(
            to_save=self._data, path=path.joinpath(f"{self._tag}.pkl"), protocol=self._protocol
        )



__all__ = [
    "BaseDataCreator",
    "HypercubeVertexDataCreator",
    "OneCacheDataCreator",
    "setup_data_creator",
]

from gravitorch.data.datacreators.base import BaseDataCreator, setup_data_creator
from gravitorch.data.datacreators.caching import OneCacheDataCreator
from gravitorch.data.datacreators.hypercube_vertex import HypercubeVertexDataCreator



__all__ = [
    "ModuleSummary",
    "compute_parameter_stats",
    "find_module_state_dict",
    "freeze_module",
    "get_module_device",
    "get_module_devices",
    "get_module_input_size",
    "get_module_name",
    "get_module_output_size",
    "has_batch_norm",
    "has_learnable_parameters",
    "has_parameters",
    "is_batch_first",
    "is_loss_decreasing",
    "is_loss_decreasing_with_adam",
    "is_loss_decreasing_with_sgd",
    "is_module_on_device",
    "load_checkpoint_to_module",
    "load_state_dict_to_module",
    "module_mode",
    "num_learnable_parameters",
    "num_parameters",
    "setup_nn_module",
    "show_parameter_stats",
    "show_state_dict_info",
    "state_dicts_are_equal",
    "top_module_mode",
    "unfreeze_module",
]

from gravitorch.nn.utils.factory import setup_nn_module
from gravitorch.nn.utils.module_helpers import (
    freeze_module,
    get_module_device,
    get_module_devices,
    get_module_input_size,
    get_module_name,
    get_module_output_size,
    has_batch_norm,
    has_learnable_parameters,
    has_parameters,
    is_batch_first,
    is_module_on_device,
    module_mode,
    num_learnable_parameters,
    num_parameters,
    top_module_mode,
    unfreeze_module,
)
from gravitorch.nn.utils.parameter_analysis import (
    compute_parameter_stats,
    show_parameter_stats,
)
from gravitorch.nn.utils.state_dict import (
    find_module_state_dict,
    load_checkpoint_to_module,
    load_state_dict_to_module,
    show_state_dict_info,
    state_dicts_are_equal,
)
from gravitorch.nn.utils.summary import ModuleSummary
from gravitorch.nn.utils.testing import (
    is_loss_decreasing,
    is_loss_decreasing_with_adam,
    is_loss_decreasing_with_sgd,
)



r"""This module defines the squared logarithmic error (SLE) metric."""

__all__ = ["SquaredAsinhError", "SquaredLogError", "SquaredSymlogError"]

import logging
from typing import Union

from torch import Tensor

from gravitorch.models.metrics.base_epoch import BaseStateEpochMetric
from gravitorch.models.metrics.state import BaseState, ErrorState
from gravitorch.nn.functional import asinh_mse_loss, msle_loss, symlog_mse_loss

logger = logging.getLogger(__name__)


class SquaredLogError(BaseStateEpochMetric):
    r"""Implements the squared logarithmic error (SLE) metric.

    This metric is best to use when targets having exponential growth,
    such as population counts, average sales of a commodity over a
    span of years etc. Note that this metric penalizes an
    under-predicted estimate greater than an over-predicted estimate.

    Note: this metric only works with positive value (0 included).

    Args:
        mode (str): Specifies the mode (e.g. train or eval).
        name (str, optional): Specifies the name of the metric. The
            name is used to log the metric results.
            Default: ``'sq_log_err'``
        state (``BaseState`` or dict, optional): Specifies the metric
            state or its configuration. If ``None``, ``ErrorState`` is
            instantiated. Default: ``None``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.models.metrics import SquaredLogError
        >>> metric = SquaredLogError('eval')
        >>> metric(torch.ones(2, 4), torch.ones(2, 4))
        >>> metric.value()
        {'eval/sq_log_err_mean': 0.0,
         'eval/sq_log_err_max': 0.0,
         'eval/sq_log_err_min': 0.0,
         'eval/sq_log_err_sum': 0.0,
         'eval/sq_log_err_num_predictions': 8}
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()
        {'eval/sq_log_err_mean': 0.08007550239562988,
         'eval/sq_log_err_max': 0.4804530143737793,
         'eval/sq_log_err_min': 0.0,
         'eval/sq_log_err_sum': 0.9609060287475586,
         'eval/sq_log_err_num_predictions': 12}
        >>> metric.reset()
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()
        {'eval/sq_log_err_mean': 0.24022650718688965,
         'eval/sq_log_err_max': 0.4804530143737793,
         'eval/sq_log_err_min': 0.0,
         'eval/sq_log_err_sum': 0.9609060287475586,
         'eval/sq_log_err_num_predictions': 4}
    """

    def __init__(
        self,
        mode: str,
        name: str = "sq_log_err",
        state: Union[BaseState, dict, None] = None,
    ):
        super().__init__(mode=mode, name=name, state=state or ErrorState())

    def forward(self, prediction: Tensor, target: Tensor) -> None:
        r"""Updates the squared logarithmic error metric given a mini-batch of
        examples.

        Args:
            prediction (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type float or long):
                Specifies the predictions.
            target (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type float or long):
                Specifies the target tensor.
        """
        self._state.update(msle_loss(prediction, target, reduction="none"))


class SquaredSymlogError(BaseStateEpochMetric):
    r"""Implements a metric to compute the squared error on the symlog
    transformed predictions and targets.

    It is a generalization of mean squared logarithmic error (MSLE)
    that works for real values. The ``symlog`` transformation is used
    instead of ``log1p`` because ``symlog`` works on negative values.

    Args:
        mode (str): Specifies the mode (e.g. train or eval).
        name (str, optional): Specifies the name of the metric. The
            name is used to log the metric results.
            Default: ``'sq_symlog_err'``
        state (``BaseState`` or dict, optional): Specifies the metric
            state or its configuration. If ``None``, ``ErrorState`` is
            instantiated. Default: ``None``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.models.metrics import SquaredSymlogError
        >>> metric = SquaredSymlogError('eval')
        >>> metric(torch.ones(2, 4), torch.ones(2, 4))
        >>> metric.value()
        {'eval/sq_symlog_err_mean': 0.0,
         'eval/sq_symlog_err_min': 0.0,
         'eval/sq_symlog_err_max': 0.0,
         'eval/sq_symlog_err_sum': 0.0,
         'eval/sq_symlog_err_num_predictions': 8}
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()
        {'eval/sq_symlog_err_mean': 0.08007550239562988,
         'eval/sq_symlog_err_min': 0.0,
         'eval/sq_symlog_err_max': 0.4804530143737793,
         'eval/sq_symlog_err_sum': 0.9609060287475586,
         'eval/sq_symlog_err_num_predictions': 12}
        >>> metric.reset()
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()
        {'eval/sq_symlog_err_mean': 0.24022650718688965,
         'eval/sq_symlog_err_min': 0.0,
         'eval/sq_symlog_err_max': 0.4804530143737793,
         'eval/sq_symlog_err_sum': 0.9609060287475586,
         'eval/sq_symlog_err_num_predictions': 4}
    """

    def __init__(
        self,
        mode: str,
        name: str = "sq_symlog_err",
        state: Union[BaseState, dict, None] = None,
    ):
        super().__init__(mode=mode, name=name, state=state or ErrorState())

    def forward(self, prediction: Tensor, target: Tensor) -> None:
        r"""Updates the squared error on the symlog transformed predictions and
        targets given a mini-batch of examples.

        Args:
            prediction (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type float or long):
                Specifies the predictions.
            target (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type float or long):
                Specifies the target tensor.
        """
        self._state.update(symlog_mse_loss(prediction, target, reduction="none"))


class SquaredAsinhError(BaseStateEpochMetric):
    r"""Implements a metric to compute the squared error on the inverse
    hyperbolic sine (arcsinh) transformed predictions and targets.

    Args:
        mode (str): Specifies the mode (e.g. train or eval).
        name (str, optional): Specifies the name of the metric. The
            name is used to log the metric results.
            Default: ``'sq_asinh_err'``
        state (``BaseState`` or dict, optional): Specifies the metric
            state or its configuration. If ``None``, ``ErrorState`` is
            instantiated. Default: ``None``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.models.metrics import SquaredAsinhError
        >>> metric = SquaredAsinhError('eval')
        >>> metric(torch.ones(2, 4), torch.ones(2, 4))
        >>> metric.value()
        {'eval/sq_asinh_err_mean': 0.0,
         'eval/sq_asinh_err_min': 0.0,
         'eval/sq_asinh_err_max': 0.0,
         'eval/sq_asinh_err_sum': 0.0,
         'eval/sq_asinh_err_num_predictions': 8}
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()
        {'eval/sq_asinh_err_mean': 0.12946990132331848,
         'eval/sq_asinh_err_min': 0.0,
         'eval/sq_asinh_err_max': 0.7768194079399109,
         'eval/sq_asinh_err_sum': 1.5536388158798218,
         'eval/sq_asinh_err_num_predictions': 12}
        >>> metric.reset()
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()
        {'eval/sq_asinh_err_mean': 0.38840970396995544,
         'eval/sq_asinh_err_min': 0.0,
         'eval/sq_asinh_err_max': 0.7768194079399109,
         'eval/sq_asinh_err_sum': 1.5536388158798218,
         'eval/sq_asinh_err_num_predictions': 4}
    """

    def __init__(
        self,
        mode: str,
        name: str = "sq_asinh_err",
        state: Union[BaseState, dict, None] = None,
    ):
        super().__init__(mode=mode, name=name, state=state or ErrorState())

    def forward(self, prediction: Tensor, target: Tensor) -> None:
        r"""Updates the squared error on the inverse hyperbolic sine (arcsinh)
        transformed predictions and targets given a mini-batch of examples.

        Args:
            prediction (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type float or long):
                Specifies the predictions.
            target (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type float or long):
                Specifies the target tensor.
        """
        self._state.update(asinh_mse_loss(prediction, target, reduction="none"))



__all__ = ["all_gather_tensor_varshape", "broadcast_object_list", "sync_reduce", "sync_reduce_"]

from typing import Optional, Union, overload

import torch
from torch import Tensor
from torch import distributed as tdist

from gravitorch.distributed.comm import all_gather, all_reduce
from gravitorch.distributed.comm import device as dist_device
from gravitorch.distributed.comm import get_world_size, is_distributed

# The supported reduction operators
AVG = "AVG"
BAND = "AND"  # Bitwise AND (only for integer/long)
BOR = "OR"  # Bitwise OR (only for integer/long)
MAX = "MAX"
MIN = "MIN"
PRODUCT = "PRODUCT"
SUM = "SUM"


def broadcast_object_list(
    object_list: list, src: int = 0, device: Optional[torch.device] = None
) -> None:
    r"""Broadcasts picklable objects in ``object_list``.

    See ``torch.distributed.broadcast_object_list`` for more
    information.

    Args:
        object_list (List[Any]): List of input objects to broadcast.
            Each object must be picklable. Only objects on the
            ``src`` rank will be broadcast, but each rank must provide
            lists of equal sizes.
        src (int, optional): Specifies the source rank from which to
            broadcast ``object_list``. Default: ``0``
        device (``torch.device``, optional): If not ``None``, the
            objects are serialized and converted to tensors which are
            moved to the ``device`` before broadcasting.
            Default is ``None``.

    Example usage:

    .. code-block:: python

        >>> # Note: Process group initialization omitted on each rank.
        >>> import torch
        >>> from gravitorch import distributed as dist
        >>> x = [10 * dist.get_local_rank(), 10 * dist.get_local_rank() + 1]
        >>> print(x)
        [0, 1]    # Rank 0
        [10, 11]  # Rank 1
        >>> from gravitorch.distributed import ddp
        >>> ddp.broadcast_object_list(x)
        >>> print(x)
        [0, 1]    # Rank 0
        [0, 1]    # Rank 1
    """
    # There is nothing to broadcast in a non-distributed environment,
    if is_distributed():
        tdist.broadcast_object_list(object_list=object_list, src=src, device=device)


@overload
def sync_reduce(variable: Tensor, op: str) -> Tensor:
    r"""``sync_reduce`` for ``torch.Tensor``."""


@overload
def sync_reduce(variable: int, op: str) -> int:
    r"""``sync_reduce`` for int."""


@overload
def sync_reduce(variable: float, op: str) -> float:
    r"""``sync_reduce`` for float."""


def sync_reduce(variable: Union[Tensor, int, float], op: str) -> Union[Tensor, int, float]:
    r"""Synchronizes all the processes and then reduce the variable.

    This function is a no-operation function if the distributed mode
    is not activated. It returns the input. If the distributed mode
    is activated, this function does not change the input variable.
    If the input is a tensor, this function will create a copy of the
    tensor before to reduce it. After this function is executed,
    the input variable will contain the value before reduction.
    If you want to do an in-place operation, you can use
    ``sync_reduce_``.

    Args:
        variable (``torch.Tensor`` or int or float): Specifies the
            variable to reduce.
        op (str): Specifies the reduction operation.
            The available operations are: ``AVG``, ``AND``, ``OR``,
            ``MAX``, ``MIN``, ``PRODUCT`` and ``SUM``.

    Returns:
        ``torch.Tensor`` or int or float: The reduced variable.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch import distributed as dist
        >>> x = torch.ones(2, 3, device=dist.device())
        >>> from gravitorch.distributed import ddp
        >>> x_reduced = ddp.sync_reduce(x, op=ddp.SUM)
    """
    if is_distributed():
        divide_by_world_size = False
        if op == AVG:
            # Average is not a supported operation by PyTorch distributed.
            op = SUM
            divide_by_world_size = True
        if torch.is_tensor(variable):
            # Create a copy to not change the values of the input tensor.
            variable = variable.clone()
        variable = all_reduce(variable, op=op)
        if divide_by_world_size:
            variable = variable / get_world_size()
    return variable


def sync_reduce_(tensor: Tensor, op: str) -> Tensor:
    r"""In-place version of ``sync_reduce`` but it works only for a tensor.

    Args:
        tensor (``torch.Tensor``): Specifies the tensor to reduce.
        op (str): Specifies the reduction operation.
            The available operations are: ``AVG``, ``AND``, ``OR``,
            ``MAX``, ``MIN``, ``PRODUCT`` and ``SUM``.

    Returns:
        The reduced tensor which is also the input tensor.

    Raises:
        ``TypeError`` if the input is not a tensor.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch import distributed as dist
        >>> x = torch.ones(2, 3, device=dist.device())
        >>> from gravitorch.distributed import ddp
        >>> ddp.sync_reduce_(x, op=ddp.SUM)  # in-place reduction
    """
    if not torch.is_tensor(tensor):
        raise TypeError("The function sync_reduce_ only supports Tensor")

    if is_distributed():
        divide_by_world_size = False
        if op == AVG:
            # Average is not a supported operation by PyTorch distributed.
            op = SUM
            divide_by_world_size = True
        all_reduce(tensor, op=op)
        if divide_by_world_size:
            tensor.div_(get_world_size())
    return tensor


def all_gather_tensor_varshape(tensor: Tensor) -> list[Tensor]:
    r"""Implements an all gather operation for variable shape tensors.

    Note: the tensors can have variable shapes, but they have to have
    the same number of dimensions. The tensor should have at least one
    dimension.

    Args:
        tensor (``torch.Tensor``): Specifies the tensor to collect
            across participating processes.

    Returns:
        list: The list of collected tensors.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> x = torch.tensor([[0, 1, 2], [3, 4, 5]])  # process 0
        >>> x = torch.tensor([[1], [0]])  # process 1
        >>> from gravitorch.distributed import ddp
        >>> ddp.all_gather_tensor_varshape(x)
        [tensor([[0, 1, 2], [3, 4, 5]]), tensor([[1], [0]])]
    """
    if not is_distributed():
        return [tensor]

    shapes = all_gather(torch.tensor(tensor.shape).unsqueeze(dim=0))
    numels = shapes.prod(dim=1)
    tensor_padded = torch.zeros(numels.max().item(), dtype=tensor.dtype, device=dist_device())
    tensor_padded[: tensor.numel()] = tensor.flatten()
    tensors_padded = all_gather(tensor_padded.unsqueeze(dim=0))
    return [values[:n].view(*shape) for n, shape, values in zip(numels, shapes, tensors_padded)]



__all__ = ["AsinhBarronRobustLoss", "BarronRobustLoss"]

from typing import Optional

from torch import Tensor
from torch.nn import Module

from gravitorch.nn.functional import asinh_barron_robust_loss, barron_robust_loss
from gravitorch.nn.functional.loss_helpers import VALID_REDUCTIONS


class BarronRobustLoss(Module):
    r"""Implements the Barron robust loss function.

    Based on the paper:

        A General and Adaptive Robust Loss Function
        Jonathan T. Barron
        CVPR 2019 (https://arxiv.org/abs/1701.03077)

    Args:
        alpha (float, optional): Specifies the shape parameter that
            controls the robustness of the loss. Default: ``2.0``
        scale (float, optional): Specifies the scale parameter that
            controls the size of the loss’s quadratic bowl near 0.
            Default: ``1.0``
        max_value (float or ``None``, optional): Specifies the max
            value to clip the loss before to compute the reduction.
            ``None`` means no clipping is used. Default: ``None``
        reduction (string, optional): Specifies the reduction to
            apply to the output: ``'none'`` | ``'mean'`` | ``'sum'``.
            ``'none'``: no reduction will be applied, ``'mean'``: the
            sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be
            summed. Default: ``"mean"``
    """

    def __init__(
        self,
        alpha: float = 2.0,
        scale: float = 1.0,
        max_value: Optional[float] = None,
        reduction: str = "mean",
    ):
        super().__init__()
        self._alpha = float(alpha)
        if scale <= 0:
            raise ValueError(f"scale has to be greater than 0 but received {scale}")
        self._scale = float(scale)
        self._max_value = max_value

        if reduction not in VALID_REDUCTIONS:
            raise ValueError(
                f"Incorrect reduction: {reduction}. The valid reductions are: {VALID_REDUCTIONS}"
            )
        self.reduction = reduction

    def extra_repr(self) -> str:
        return (
            f"alpha={self._alpha}, scale={self._scale}, max_value={self._max_value}, "
            f"reduction={self.reduction}"
        )

    def forward(self, prediction: Tensor, target: Tensor) -> Tensor:
        r"""Computes the loss values and reduces them.

        Args:
            prediction (``torch.Tensor``): Specifies the predictions.
            target (``torch.Tensor``): Specifies the targets.

        Returns:
            ``torch.Tensor``: The computed loss value.
        """
        return barron_robust_loss(
            prediction=prediction,
            target=target,
            alpha=self._alpha,
            scale=self._scale,
            max_value=self._max_value,
            reduction=self.reduction,
        )


class AsinhBarronRobustLoss(BarronRobustLoss):
    r"""Implements the Barron robust loss function where the inverse hyperbolic
    sine (arcsinh) transformation is applied on the prediction and target."""

    def forward(self, prediction: Tensor, target: Tensor) -> Tensor:
        r"""Computes the loss values and reduces them.

        Args:
            prediction (``torch.Tensor``): Specifies the predictions.
            target (``torch.Tensor``): Specifies the targets.

        Returns:
            ``torch.Tensor``: The computed loss value.
        """
        return asinh_barron_robust_loss(
            prediction=prediction,
            target=target,
            alpha=self._alpha,
            scale=self._scale,
            max_value=self._max_value,
            reduction=self.reduction,
        )



from __future__ import annotations

__all__ = ["BaseOptimizerCreator"]

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from objectory import AbstractFactory
from torch.nn import Module
from torch.optim import Optimizer

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine


class BaseOptimizerCreator(ABC, metaclass=AbstractFactory):
    r"""Defines the base class to create an optimizer.

    Note that it is not the unique approach to create an optimizer. Feel
    free to use other approaches if this approach does not fit your
    needs.
    """

    @abstractmethod
    def create(self, engine: BaseEngine, model: Module) -> Optimizer | None:
        r"""Creates an optimizer.

        This method is responsible to register the event handlers
        associated to the optimizer.

        Args:
            engine (``gravitorch.engines.BaseEngine``): Specifies an
                engine.
            model (``torch.nn.Module``): Specifies a model.

        Returns:
            ``torch.optim.Optimizer`` or ``None``: The created
                optimizer or ``None`` if no optimizer is created.
        """



__all__ = ["ScalarMeter"]

from collections import deque
from collections.abc import Iterable
from typing import Any, Union

import torch

from gravitorch.utils.format import to_pretty_dict_str
from gravitorch.utils.meters.exceptions import EmptyMeterError


class ScalarMeter:
    r"""Implementation of a scalar meter.

    This meter tracks the following values:

        - the sum of the values
        - the number of values
        - the minimum value
        - the maximum value
        - the last N values which are used to compute the median value.

    Args:
        total (float, optional): Specifies the initial total value.
            Default: ``0.0``
        count (int, optional): Specifies the initial count value.
            Default: ``0``
        min_value (float, optional): Specifies the initial minimum
            value. Default: ``inf``
        max_value (float, optional): Specifies the initial maximum
            value. Default: ``-inf``
        values (iterable, optional): Specifies the initial last
            values to store.
        max_size (int, optional): Specifies the maximum size used to
            store the last values because it may not be possible to
            store all the values. This parameter is used to compute
            the median only on the N last values. Default: ``100``

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.meters import ScalarMeter
        >>> meter = ScalarMeter()
        >>> meter.update(6)
        >>> meter.update_sequence([1, 2, 3, 4, 5, 0])
        >>> print(meter)
        ScalarMeter
          average : 3.0
          count   : 7
          max     : 6
          median  : 3
          min     : 0
          std     : 2.1602468490600586
          sum     : 21.0
        >>> meter.average()
        3.0
    """

    def __init__(
        self,
        total: float = 0.0,
        count: int = 0,
        min_value: float = float("inf"),
        max_value: float = -float("inf"),
        values: Iterable[float] = tuple(),
        max_size: int = 100,
    ):
        self._total = float(total)
        self._count = int(count)
        self._min_value = float(min_value)
        self._max_value = float(max_value)
        # Store only the N last values to scale to large number of values.
        self._values = deque(values, maxlen=max_size)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(count={self._count:,}, total={self._total}, "
            f"min_value={self._min_value}, max_value={self._max_value}, "
            f"max_size={self._values.maxlen:,})"
        )

    def __str__(self) -> str:
        count = self.count
        stats = to_pretty_dict_str(
            {
                "average": self.average() if count else "N/A (empty)",
                "count": count,
                "max": self.max() if count else "N/A (empty)",
                "median": self.median() if count else "N/A (empty)",
                "min": self.min() if count else "N/A (empty)",
                "std": self.std() if count else "N/A (empty)",
                "sum": self.sum() if count else "N/A (empty)",
            },
            indent=2,
        )
        return f"{self.__class__.__qualname__}\n{stats}"

    @property
    def count(self) -> int:
        r"""int: The number of examples in the meter."""
        return self._count

    @property
    def total(self) -> float:
        r"""float: The total of the values added to the meter."""
        return self._total

    @property
    def values(self) -> tuple[float, ...]:
        r"""tuple: The values store in this meter. If there are more
        values that the maximum size, only the last values are
        returned.
        """
        return tuple(self._values)

    def average(self) -> float:
        r"""Computes the average value.

        Returns:
            float: The average value.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        if not self._count:
            raise EmptyMeterError("The meter is empty")
        return self._total / float(self._count)

    def equal(self, other) -> bool:
        r"""Indicates if two meters are equal or not.

        Args:
            other: Specifies the value to compare.

        Returns:
            bool: ``True`` if the meters are equal,
                ``False`` otherwise.
        """
        if not isinstance(other, ScalarMeter):
            return False
        return self.state_dict() == other.state_dict()

    def load_state_dict(self, state_dict: dict[str, Any]) -> None:
        r"""Loads a state to the history tracker.

        Args:
            state_dict (dict): Dictionary containing state keys with
                values.
        """
        self._total = state_dict["total"]
        self._count = state_dict["count"]
        self._max_value = state_dict["max_value"]
        self._min_value = state_dict["min_value"]
        self._values.clear()
        self._values.extend(state_dict["values"])

    def max(self) -> float:
        r"""Gets the max value.

        Returns:
            float: The max value.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        if not self._count:
            raise EmptyMeterError("The meter is empty")
        return self._max_value

    def median(self) -> float:
        r"""Computes the median value from the last examples.

        If there are more values than the maximum window size, only
        the last examples are used. Internally, this meter uses a
        deque to track the last values and the median value is
        computed on the values in the deque. The median is not unique
        for input tensors with an even number of elements. In this
        case the lower of the two medians is returned.

        Returns:
            float: The median value from the last examples.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        if not self._count:
            raise EmptyMeterError("The meter is empty")
        return torch.tensor(list(self._values)).median().item()

    def merge(self, meters: Iterable["ScalarMeter"]) -> "ScalarMeter":
        r"""Merges several meters with the current meter and returns a new
        meter.

        Only the values of the current meter are copied to the merged
        meter.

        Args:
            meters (iterable): Specifies the meters to merge to the
                current meter.

        Returns:
            ``AverageMeter``: The merged meter.
        """
        count, total = self._count, self._total
        min_value, max_value = self._min_value, self._max_value
        for meter in meters:
            count += meter.count
            total += meter.total
            min_value = min(min_value, meter._min_value)
            max_value = max(max_value, meter._max_value)
        return ScalarMeter(
            total=total,
            count=count,
            min_value=min_value,
            max_value=max_value,
            values=self.values,
            max_size=self._values.maxlen,
        )

    def merge_(self, meters: Iterable["ScalarMeter"]) -> None:
        r"""Merges several meters into the current meter.

        In-place version of ``merge``.

        Only the values of the current meter are copied to the merged
        meter.

        Args:
            meters (iterable): Specifies the meters to merge to the
                current meter.

        Returns:
            ``AverageMeter``: The merged meter.
        """
        for meter in meters:
            self._count += meter.count
            self._total += meter.total
            self._min_value = min(self._min_value, meter._min_value)
            self._max_value = max(self._max_value, meter._max_value)

    def min(self) -> float:
        r"""Gets the min value.

        Returns:
            float: The min value.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        if not self._count:
            raise EmptyMeterError("The meter is empty")
        return self._min_value

    def reset(self) -> None:
        r"""Reset the meter."""
        self._total = 0.0
        self._count = 0
        self._min_value = float("inf")
        self._max_value = -float("inf")
        self._values.clear()

    def state_dict(self) -> dict[str, Any]:
        r"""Returns a dictionary containing state values.

        Returns:
            dict: The state values in a dict.
        """
        return {
            "count": self._count,
            "total": self._total,
            "values": tuple(self._values),
            "max_value": self._max_value,
            "min_value": self._min_value,
        }

    def std(self) -> float:
        r"""Computes the standard deviation from the last examples.

        If there are more values than the maximum window size, only
        the last examples are used. Internally, this meter uses a
        deque to track the last values and the standard deviation
        is computed on the values in the deque.

        Returns:
            float: The standard deviation from the last examples.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        if not self._count:
            raise EmptyMeterError("The meter is empty")
        return torch.tensor(self.values, dtype=torch.float).std(dim=0).item()

    def sum(self) -> float:
        r"""Computes the sum value.

        Returns:
            float: The sum value.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        if not self._count:
            raise EmptyMeterError("The meter is empty")
        return self._total

    def update(self, value: float) -> None:
        r"""Updates the meter given a new value.

        Args:
            value (float): Specifies the value to add to the meter.
        """
        value = float(value)
        self._total += value
        self._count += 1
        self._min_value = min(self._min_value, value)
        self._max_value = max(self._max_value, value)
        self._values.append(value)

    def update_sequence(self, values: Union[list[float], tuple[float, ...]]) -> None:
        r"""Updates the meter given a list/tuple of values.

        Args:
            values (list or tuple): Specifies the list/tuple of
                values to add to the meter.
        """
        self._total += float(sum(values))
        self._count += len(values)
        self._min_value = float(min(self._min_value, *values))
        self._max_value = float(max(self._max_value, *values))
        self._values.extend(float(v) for v in values)



__all__ = ["NoOptimizerCreator"]

import logging

from torch.nn import Module

from gravitorch.creators.optimizer.base import BaseOptimizerCreator
from gravitorch.engines.base import BaseEngine

logger = logging.getLogger(__name__)


class NoOptimizerCreator(BaseOptimizerCreator):
    r"""Implements a no optimizer creator.

    This optimizer creator should be used if you do not want to create
    an optimizer. For example if you only want to evaluate your model,
    you do not need to create an optimizer. The ``create`` method always
    returns ``None``.
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def create(self, engine: BaseEngine, model: Module) -> None:
        r"""Does not create an optimizer.

        Args:
            engine (``gravitorch.engines.BaseEngine``): Specifies an
                engine.
            model (``torch.nn.Module``): Specifies a model.

        Returns:
            ``None``: because there is no optimizer to create.
        """
        logger.info("No optimizer")
        return None



__all__ = ["BaseHandler"]

from abc import ABC, abstractmethod

from objectory import AbstractFactory

from gravitorch.engines.base import BaseEngine


class BaseHandler(ABC, metaclass=AbstractFactory):
    r"""Define the base class for the handlers."""

    @abstractmethod
    def attach(self, engine: BaseEngine) -> None:
        r"""Attaches the handler to the engine.

        Args:
            engine (``BaseEngine``): Specifies the engine used to
                attach the handler.
        """



__all__ = ["FusionNorm"]

from typing import Any, Union

from torch import Tensor
from torch.nn import Dropout, LayerNorm, Module

from gravitorch.nn.fusion.multiplication import MultiplicationFusion
from gravitorch.nn.fusion.sum import SumFusion
from gravitorch.nn.utils import setup_nn_module


class FusionNorm(Module):
    """Implements a layer that fuses multiple inputs and then computes a
    normalization.

    Args:
        fusion (``torch.nn.Module`` or dict): Specifies the fusion
            layer or its configuration.
        norm (``torch.nn.Module`` or dict): Specifies the
            normalization layer or its configuration.
        dropout (``torch.nn.Module`` or dict): Specifies the dropout
            layer or its configuration.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.nn import FusionNorm
        >>> x1 = torch.tensor([[2, 3, 4], [5, 6, 7]], dtype=torch.float, requires_grad=True)
        >>> x2 = torch.tensor([[12, 13, 14], [15, 16, 17]], dtype=torch.float, requires_grad=True)
        >>> module = FusionNorm.create_sum_layer_norm(3)
        >>> module
        FusionNorm(
          (fusion): SumFusion(normalized=False)
          (norm): LayerNorm((3,), eps=1e-06, elementwise_affine=True)
          (dropout): Dropout(p=0.1, inplace=False)
        )
        >>> out = module(x1, x2)
        >>> out
        tensor([[-1.3608,  0.0000,  1.3608],
                [-0.0000,  0.0000,  1.3608]], grad_fn=<MulBackward0>)
        >>> out.mean().backward()
        >>> module = FusionNorm.create_multiplication_layer_norm(3)
        >>> module
        FusionNorm(
          (fusion): MultiplicationFusion()
          (norm): LayerNorm((3,), eps=1e-06, elementwise_affine=True)
          (dropout): Dropout(p=0.1, inplace=False)
        )
        >>> out = module(x1, x2)
        >>> out
        tensor([[-1.3316, -0.0567,  1.3883],
                [-1.3397, -0.0412,  1.3810]], grad_fn=<MulBackward0>)
        >>> out.mean().backward()
    """

    def __init__(
        self,
        fusion: Union[Module, dict[str, Any]],
        norm: Union[Module, dict[str, Any]],
        dropout: Union[Module, dict[str, Any]],
    ):
        super().__init__()
        self.fusion = setup_nn_module(fusion)
        self.norm = setup_nn_module(norm)
        self.dropout = setup_nn_module(dropout)

    def forward(self, *inputs: Tensor) -> Tensor:
        r"""Fuses a list or tuple of inputs.

        Args:
            *inputs (list or tuple of tensors): Specifies the list or
                tuple of tensors to fuse. The shape of the tensors
                should be the same.

        Returns:
            ``torch.Tensor`` with the same shape that the input
                tensors: The fused tensor.
        """
        return self.dropout(self.norm(self.fusion(*inputs)))

    @classmethod
    def create_sum_layer_norm(
        cls,
        input_size: int,
        layer_norm_eps: float = 1e-6,
        dropout: float = 0.1,
    ) -> "FusionNorm":
        r"""Instantiates a ``FusionNorm`` layer with a sum fusion and a layer
        norm.

        Args:
            input_size (int): Specifies the input size.
            layer_norm_eps (float, optional): Specifies the ``eps``
                value in the ``LayerNorm`` layer. Default: ``1e-6``
            dropout (float, optional): Specifies the dropout value.
                Default: ``0.1``

        Returns:
            ``FusionNorm``: A ``FusionNorm`` layer with a sum fusion
                and a layer norm.
        """
        return cls(
            fusion=SumFusion(),
            norm=LayerNorm(input_size, eps=layer_norm_eps),
            dropout=Dropout(dropout),
        )

    @classmethod
    def create_multiplication_layer_norm(
        cls,
        input_size: int,
        layer_norm_eps: float = 1e-6,
        dropout: float = 0.1,
    ) -> "FusionNorm":
        r"""Instantiates a ``FusionNorm`` layer with a multiplication fusion and
        a layer norm.

        Args:
            input_size (int): Specifies the input size.
            layer_norm_eps (float, optional): Specifies the ``eps``
                value in the ``LayerNorm`` layer. Default: ``1e-6``
            dropout (float, optional): Specifies the dropout value.
                Default: ``0.1``

        Returns:
            ``FusionNorm``: A ``FusionNorm`` layer with a
                multiplication fusion and a layer norm.
        """
        return cls(
            fusion=MultiplicationFusion(),
            norm=LayerNorm(input_size, eps=layer_norm_eps),
            dropout=Dropout(dropout),
        )



__all__ = [
    "EngineStateLoader",
    "EngineStateLoaderWithExcludeKeys",
    "EngineStateLoaderWithIncludeKeys",
]

import copy
import logging
from pathlib import Path
from typing import Union

import torch

from gravitorch.engines.base import BaseEngine
from gravitorch.handlers.base import BaseHandler
from gravitorch.handlers.utils import add_unique_event_handler
from gravitorch.utils.events import VanillaEventHandler
from gravitorch.utils.path import sanitize_path

logger = logging.getLogger(__name__)


class EngineStateLoader(BaseHandler):
    r"""Implements a handler to load the engine state dict from a PyTorch file.

    Note that the state dict in the file should be compatible with
    the engine.

    Args:
        path (``pathlib.Path`` or str): Specifies the path to the
            PyTorch file.
        event (str): Specifies the event used to load the engine
            state dict.
        missing_ok (bool, optional): If ``False``, a
            ``FileNotFoundError`` exception is raised if the path
            does not exist, otherwise the ``FileNotFoundError``
            exception is not raised and this handler does not try
            to load the state dict. Default: ``False``
    """

    def __init__(self, path: Union[Path, str], event: str, missing_ok: bool = False):
        self._path = sanitize_path(path)
        self._event = str(event)
        self._missing_ok = bool(missing_ok)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(event={self._event}, path={self._path}, "
            f"missing_ok={self._missing_ok})"
        )

    def attach(self, engine: BaseEngine) -> None:
        add_unique_event_handler(
            engine=engine,
            event=self._event,
            event_handler=VanillaEventHandler(
                self.load_engine_state_dict, handler_kwargs={"engine": engine}
            ),
        )

    def load_engine_state_dict(self, engine: BaseEngine) -> None:
        r"""Loads an engine state dict from a PyTorch file.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        if not self._path.is_file():
            if self._missing_ok:
                return
            else:
                raise FileNotFoundError(f"{self._path} is not a file")
        logger.info(f"Load state dict from {self._path}...")
        state = torch.load(self._path, map_location=lambda storage, loc: storage)
        engine.load_state_dict(self._prepare_state_dict(state))

    def _prepare_state_dict(self, state_dict: dict) -> dict:
        r"""Prepares the state dict before to load it in the engine.

        Args:
            state_dict (dict): The state dict to prepare.

        Returns:
            dict: The prepared state dict.
        """
        return state_dict


class EngineStateLoaderWithExcludeKeys(EngineStateLoader):
    r"""Implements a handler to load the engine state dict from a PyTorch file.

    This state dict loader excludes the keys that are specified in
    ``exclude_keys``. Note that the state dict in the file should be
    compatible with the engine.

    Args:
        path (``pathlib.Path`` or str): Specifies the path to the
            PyTorch file.
        event (str): Specifies the event used to load the engine
            state dict.
        exclude_keys (tuple or list): Specifies the keys to exclude
            in the state dict.
        missing_ok (bool, optional): If ``False``, a
            ``FileNotFoundError`` exception is raised if the path
            does not exist, otherwise the ``FileNotFoundError``
            exception is not raised and this handler does not try
            to load the state dict. Default: ``False``
    """

    def __init__(
        self,
        path: Union[Path, str],
        event: str,
        exclude_keys: Union[tuple[str, ...], list[str]],
        missing_ok: bool = False,
    ):
        super().__init__(path=path, event=event, missing_ok=missing_ok)
        self._exclude_keys = tuple(exclude_keys)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(event={self._event}, path={self._path}, "
            f"exclude_keys={self._exclude_keys}, missing_ok={self._missing_ok})"
        )

    def _prepare_state_dict(self, state_dict: dict) -> dict:
        r"""Prepares the state dict before to load it in the engine.

        Args:
            state_dict (dict): The state dict to prepare.

        Returns:
            dict: The prepared state dict.
        """
        new_state_dict = {}
        logger.info(
            "Before preparation, the state dict has the following keys: "
            f"{tuple(state_dict.keys())}"
        )
        for key, value in state_dict.items():
            if key not in self._exclude_keys:
                new_state_dict[key] = copy.copy(value)
        logger.info(
            "After preparation, the state dict has the following keys: "
            f"{tuple(new_state_dict.keys())}"
        )
        return new_state_dict


class EngineStateLoaderWithIncludeKeys(EngineStateLoader):
    r"""Implements a handler to load the engine state dict from a PyTorch file.

    This state dict loader loads only the keys that are specified in
    ``include_keys``. If a key is not present in the state dict, this
    key is ignored. Note that the state dict in the file should be
    compatible with the engine.

    Args:
        path (``pathlib.Path`` or str): Specifies the path to the
            PyTorch file.
        event (str): Specifies the event used to load the engine state
            dict.
        include_keys (tuple or list): Specifies the keys to include in
            the state dict. If a key is in the state dict but not in
            ``include_keys``, this key is ignored.
        missing_ok (bool, optional): If ``False``, a
            ``FileNotFoundError`` exception is raised if the path
            does not exist, otherwise the ``FileNotFoundError``
            exception is not raised and this handler does not try to
            load the state dict. Default: ``False``
    """

    def __init__(
        self,
        path: Union[Path, str],
        event: str,
        include_keys: Union[tuple[str, ...], list[str]],
        missing_ok: bool = False,
    ):
        super().__init__(path=path, event=event, missing_ok=missing_ok)
        self._include_keys = tuple(include_keys)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(event={self._event}, path={self._path}, "
            f"include_keys={self._include_keys}, missing_ok={self._missing_ok})"
        )

    def _prepare_state_dict(self, state_dict: dict) -> dict:
        r"""Prepares the state dict before to load it in the engine.

        Args:
            state_dict (dict): The state dict to prepare.

        Returns:
            dict: The prepared state dict.
        """
        new_state_dict = {}
        logger.info(
            f"Before preparation, the state dict has the following keys: {tuple(state_dict.keys())}"
        )
        for key in self._include_keys:
            if key in state_dict:
                new_state_dict[key] = copy.copy(state_dict[key])
            else:
                logger.info(f"Ignore '{key}' key because it is not present in the state dict")
        logger.info(
            "After preparation, the state dict has the following keys: "
            f"{tuple(new_state_dict.keys())}"
        )
        return new_state_dict



__all__ = ["BaseRunner"]

from abc import ABC, abstractmethod
from typing import Any

from objectory import AbstractFactory


class BaseRunner(ABC, metaclass=AbstractFactory):
    r"""Defines the base class of the runners."""

    @abstractmethod
    def run(self) -> Any:
        r"""Executes the logic of the runner.

        Returns:
            Any artifact of the runner
        """



__all__ = [
    "PyTorchConfig",
    "PyTorchCudaBackend",
    "PyTorchCudaBackendState",
    "PyTorchCudnnBackend",
    "PyTorchCudnnBackendState",
]

import logging
from dataclasses import dataclass
from types import TracebackType
from typing import Any, Optional

import torch
from torch.backends import cuda, cudnn

from gravitorch.rsrc.base import BaseResource
from gravitorch.utils.format import to_pretty_dict_str

logger = logging.getLogger(__name__)


class PyTorchConfig(BaseResource):
    r"""Implements a context manager to show the PyTorch configuration."""

    def __enter__(self) -> "PyTorchConfig":
        logger.info(f"PyTorch version: {torch.version.__version__}  ({torch.version.git_version})")
        logger.info(f"PyTorch cuda version: {torch.version.cuda}")
        logger.info(f"PyTorch configuration:\n{torch.__config__.show()}")
        logger.info(f"PyTorch parallel information:\n{torch.__config__.parallel_info()}")
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        pass

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"


@dataclass
class PyTorchCudaBackendState:
    allow_tf32: bool
    allow_fp16_reduced_precision_reduction: bool
    flash_sdp_enabled: bool
    math_sdp_enabled: bool
    preferred_linalg_backend: Any

    def restore(self) -> None:
        r"""Restores the PyTorch CUDA backend configuration by using the values
        in the state."""
        cuda.matmul.allow_tf32 = self.allow_tf32
        cuda.matmul.allow_fp16_reduced_precision_reduction = (
            self.allow_fp16_reduced_precision_reduction
        )
        cuda.enable_math_sdp(self.math_sdp_enabled)
        cuda.enable_flash_sdp(self.flash_sdp_enabled)
        cuda.preferred_linalg_library(self.preferred_linalg_backend)

    @classmethod
    def create(cls) -> "PyTorchCudaBackendState":
        r"""Creates a state to capture the current PyTorch CUDA backend.

        Returns:
            ``PyTorchCudaBackendState``: The current state.
        """
        return cls(
            allow_tf32=cuda.matmul.allow_tf32,
            allow_fp16_reduced_precision_reduction=(
                cuda.matmul.allow_fp16_reduced_precision_reduction
            ),
            math_sdp_enabled=cuda.math_sdp_enabled(),
            flash_sdp_enabled=cuda.flash_sdp_enabled(),
            preferred_linalg_backend=cuda.preferred_linalg_library(),
        )


class PyTorchCudaBackend(BaseResource):
    r"""Implements a context manager to configure the PyTorch CUDA backend.

    Args:
        allow_tf32 (bool or ``None``, optional): Specifies the value
            of ``torch.backends.cuda.matmul.allow_tf32``.
            If ``None``, the default value is used. Default: ``None``
        allow_fp16_reduced_precision_reduction (bool or ``None``,
            optional): Specifies the value of
            ``torch.backends.cuda.matmul.allow_fp16_reduced_precision_reduction``.
            If ``None``, the default value is used. Default: ``None``
        flash_sdp_enabled (bool or ``None``, optional): Specifies the
            value  of ``torch.backends.cuda.flash_sdp_enabled``.
            If ``None``, the default value is used. Default: ``None``
        math_sdp_enabled (bool or ``None``, optional): Specifies the
            value of ``torch.backends.cuda.math_sdp_enabled``.
            If ``None``, the default value is used. Default: ``None``
        preferred_linalg_backend (str or ``None``, optional):
            Specifies the value of
            ``torch.backends.cuda.preferred_linalg_library``.
            If ``None``, the default value is used. Default: ``None``
        log_info (bool, optional): If ``True``, the state is shown
            after the context manager is created. Default: ``False``
    """

    def __init__(
        self,
        allow_tf32: Optional[bool] = None,
        allow_fp16_reduced_precision_reduction: Optional[bool] = None,
        flash_sdp_enabled: Optional[bool] = None,
        math_sdp_enabled: Optional[bool] = None,
        preferred_linalg_backend: Optional[str] = None,
        log_info: bool = False,
    ):
        self._allow_tf32 = allow_tf32
        self._allow_fp16_reduced_precision_reduction = allow_fp16_reduced_precision_reduction
        self._flash_sdp_enabled = flash_sdp_enabled
        self._math_sdp_enabled = math_sdp_enabled
        self._preferred_linalg_backend = preferred_linalg_backend

        self._log_info = bool(log_info)
        self._state: list[PyTorchCudaBackendState] = []

    def __enter__(self) -> "PyTorchCudaBackend":
        logger.info("Configuring CUDA backend...")
        self._state.append(PyTorchCudaBackendState.create())
        self._configure()
        if self._log_info:
            self._show()
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        logger.info("Restoring CUDA backend configuration...")
        self._state.pop().restore()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(allow_tf32={self._allow_tf32}, "
            "allow_fp16_reduced_precision_reduction="
            f"{self._allow_fp16_reduced_precision_reduction}, "
            f"flash_sdp_enabled={self._flash_sdp_enabled}, "
            f"math_sdp_enabled={self._math_sdp_enabled}, "
            f"preferred_linalg_backend={self._preferred_linalg_backend}, "
            f"log_info={self._log_info})"
        )

    def _configure(self) -> None:
        if self._allow_tf32 is not None:
            cuda.matmul.allow_tf32 = self._allow_tf32
        if self._allow_fp16_reduced_precision_reduction is not None:
            cuda.matmul.allow_fp16_reduced_precision_reduction = (
                self._allow_fp16_reduced_precision_reduction
            )
        if self._flash_sdp_enabled is not None:
            cuda.enable_flash_sdp(self._flash_sdp_enabled)
        if self._math_sdp_enabled is not None:
            cuda.enable_math_sdp(self._math_sdp_enabled)
        if self._preferred_linalg_backend is not None:
            cuda.preferred_linalg_library(self._preferred_linalg_backend)

    def _show(self) -> None:
        prefix = "torch.backends.cuda"
        info = {
            f"{prefix}.matmul.allow_fp16_reduced_precision_reduction": (
                cuda.matmul.allow_fp16_reduced_precision_reduction
            ),
            f"{prefix}.matmul.allow_tf32": cuda.matmul.allow_tf32,
            f"{prefix}.is_built": cuda.is_built(),
            f"{prefix}.flash_sdp_enabled": cuda.flash_sdp_enabled(),
            f"{prefix}.math_sdp_enabled": cuda.math_sdp_enabled(),
            f"{prefix}.preferred_linalg_library": cuda.preferred_linalg_library(),
            "torch.version.cuda": torch.version.cuda,
        }
        logger.info(f"CUDA backend:\n{to_pretty_dict_str(info, sorted_keys=True, indent=2)}\n")


@dataclass
class PyTorchCudnnBackendState:
    allow_tf32: bool
    benchmark: bool
    benchmark_limit: Optional[int]
    deterministic: bool
    enabled: bool

    def restore(self) -> None:
        r"""Restores the PyTorch CUDNN backend configuration by using the
        values in the state."""
        cudnn.allow_tf32 = self.allow_tf32
        cudnn.benchmark = self.benchmark
        cudnn.benchmark_limit = self.benchmark_limit
        cudnn.deterministic = self.deterministic
        cudnn.enabled = self.enabled

    @classmethod
    def create(cls) -> "PyTorchCudnnBackendState":
        r"""Creates a state to capture the current PyTorch CUDA CUDNN.

        Returns:
            ``PyTorchCudnnBackendState``: The current state.
        """
        return cls(
            allow_tf32=cudnn.allow_tf32,
            benchmark=cudnn.benchmark,
            benchmark_limit=cudnn.benchmark_limit,
            deterministic=cudnn.deterministic,
            enabled=cudnn.enabled,
        )


class PyTorchCudnnBackend(BaseResource):
    r"""Implements a context manager to configure the PyTorch CUDNN backend.

    Args:
        allow_tf32 (bool or ``None``, optional): Specifies the value
            of ``torch.backends.cudnn.allow_tf32``. If ``None``,
            the default value is used. Default: ``None``
        benchmark (bool or ``None``, optional): Specifies the value of
            ``torch.backends.cudnn.benchmark``. If ``None``,
            the default value is used. Default: ``None``
        benchmark_limit (int or ``None``, optional): Specifies the
            value of ``torch.backends.cudnn.benchmark_limit``.
            If ``None``, the default value is used. Default: ``None``
        deterministic (bool or ``None``, optional): Specifies the
            value of ``torch.backends.cudnn.deterministic``.
            If ``None``, the default value is used. Default: ``None``
        enabled (bool or ``None``, optional): Specifies the value of
            ``torch.backends.cudnn.enabled``. If ``None``,
            the default value is used. Default: ``None``
        log_info (bool, optional): If ``True``, the state is shown
            after the context manager is created. Default: ``False``
    """

    def __init__(
        self,
        allow_tf32: bool = None,
        benchmark: Optional[bool] = None,
        benchmark_limit: Optional[int] = None,
        deterministic: Optional[bool] = None,
        enabled: Optional[bool] = None,
        log_info: bool = False,
    ):
        self._allow_tf32 = allow_tf32
        self._benchmark = benchmark
        self._benchmark_limit = benchmark_limit
        self._deterministic = deterministic
        self._enabled = enabled

        self._log_info = bool(log_info)
        self._state: list[PyTorchCudnnBackendState] = []

    def __enter__(self) -> "PyTorchCudnnBackend":
        logger.info("Configuring CUDNN backend...")
        self._state.append(PyTorchCudnnBackendState.create())
        self._configure()
        if self._log_info:
            self._show()
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        logger.info("Restoring CUDNN backend configuration...")
        self._state.pop().restore()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(allow_tf32={self._allow_tf32}, "
            f"benchmark={self._benchmark}, benchmark_limit={self._benchmark_limit}, "
            f"deterministic={self._deterministic}, enabled={self._enabled}, "
            f"log_info={self._log_info})"
        )

    def _configure(self) -> None:
        if self._allow_tf32 is not None:
            cudnn.allow_tf32 = self._allow_tf32
        if self._benchmark is not None:
            cudnn.benchmark = self._benchmark
        if self._benchmark_limit is not None:
            cudnn.benchmark_limit = self._benchmark_limit
        if self._deterministic is not None:
            cudnn.deterministic = self._deterministic
        if self._enabled is not None:
            cudnn.enabled = self._enabled

    def _show(self) -> None:
        prefix = "torch.backends.cudnn"
        info = {
            f"{prefix}.allow_tf32": cudnn.allow_tf32,
            f"{prefix}.benchmark": cudnn.benchmark,
            f"{prefix}.benchmark_limit": cudnn.benchmark_limit,
            f"{prefix}.deterministic": cudnn.deterministic,
            f"{prefix}.enabled": cudnn.enabled,
            f"{prefix}.is_available": cudnn.is_available(),
            f"{prefix}.version": cudnn.version(),
        }
        logger.info(f"CUDNN backend:\n{to_pretty_dict_str(info, sorted_keys=True, indent=2)}\n")



__all__ = ["EvenPartitioner"]

from collections.abc import Sequence
from typing import Optional, TypeVar

from gravitorch.data.partitioners.base import BasePartitioner
from gravitorch.engines import BaseEngine
from gravitorch.utils.partitioning import even_partitions

T = TypeVar("T")


class EvenPartitioner(BasePartitioner[T]):
    r"""Implements a partitioner that creates even partitions i.e. partitions
    with (almost) equal number of items.

    Args:
        num_partitions (int): Specifies the number of partitions.
        drop_remainder (bool, optional): If ``True``, it drops the
            last items if the number of items is not evenly divisible
            by ``num_partitions``.
    """

    def __init__(self, num_partitions: int, drop_remainder: bool = False):
        self._num_partitions = int(num_partitions)
        self._drop_remainder = bool(drop_remainder)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(num_partitions={self._num_partitions:,}, "
            f"drop_remainder={self._drop_remainder})"
        )

    @property
    def drop_remainder(self) -> bool:
        r"""bool: Indicates if the last items are dropped or not if
        the number of items is not evenly divisible by ``num_partitions``.
        """
        return self._drop_remainder

    @property
    def num_partitions(self) -> int:
        r"""int: The number of partitions."""
        return self._num_partitions

    def partition(
        self, items: Sequence[T], engine: Optional[BaseEngine] = None
    ) -> list[Sequence[T]]:
        return even_partitions(
            items=items,
            num_partitions=self._num_partitions,
            drop_remainder=self._drop_remainder,
        )



__all__ = ["PyTorchMnistNet"]

from collections import OrderedDict

import torch
from torch import Tensor, nn

from gravitorch import constants as ct
from gravitorch.nn.utils.module_helpers import get_module_device


class PyTorchMnistNet(nn.Module):
    r"""Implements the network architecture from PyTorch example.

    https://github.com/pytorch/examples/blob/master/mnist/main.py

    Args:
        num_classes (int, optional): Specifies the number of classes.
            Default: 10
        input_name (str, optional): Specifies the name of the input.
            Default: ``'input'``
        output_name (str, optional): Specifies the name of the output.
            Default: ``'prediction'``
    """

    def __init__(
        self, num_classes: int = 10, input_name: str = ct.INPUT, output_name: str = ct.PREDICTION
    ):
        super().__init__()
        self.net = nn.Sequential(
            OrderedDict(
                [
                    ("conv1", nn.Conv2d(1, 32, 3, 1)),
                    ("relu1", nn.ReLU()),
                    ("conv2", nn.Conv2d(32, 64, 3, 1)),
                    ("relu2", nn.ReLU()),
                    ("pool", nn.MaxPool2d(kernel_size=2)),
                    ("dropout1", nn.Dropout(0.25)),
                    ("flatten", nn.Flatten()),
                    ("fc1", nn.Linear(9216, 128)),
                    ("relu3", nn.ReLU()),
                    ("dropout2", nn.Dropout(0.5)),
                    ("fc2", nn.Linear(128, int(num_classes))),
                ]
            )
        )
        self._input_name = str(input_name)
        self._output_name = str(output_name)

    def forward(self, tensor: Tensor) -> Tensor:
        r"""Computes the predictions.

        Args:
            tensor (``torch.Tensor`` of shape
                ``(batch size, 1, 28, 28)`` and type float):
                Specifies the input images.

        Returns:
            ``torch.Tensor`` of shape ``(batch size, num classes)``
                and type float: The logits for each class.
        """
        return self.net(tensor)

    def get_dummy_input(self, batch_size: int = 1) -> tuple[Tensor]:
        r"""Generates a dummy input for the MLP.

        Args:
            batch_size (int, optional): Specifies the batch size to
                use to generate the dummy input. Default: ``1``

        Returns:
            ``tuple[Tensor]``: A tuple with one tensor of shape
                ``(batch_size, 1, 28, 28)`` and type ``float``.
                The tensor is on the same device that this network.
        """
        return (torch.rand(batch_size, 1, 28, 28, device=get_module_device(self.net)),)

    def get_input_names(self) -> tuple[str]:
        r"""Gets the tuple of input names for the network.

        The order of the name should be the same that the order in the
        inputs of the forward function.

        Returns:
            tuple: The tuple of input names.
        """
        return (self._input_name,)

    def get_onnx_dynamic_axis(self) -> dict:
        r"""Gets the dynamic axes (e.g. batch size or sequence length) when
        serializing a model to ONNX format.

        See https://pytorch.org/docs/stable/onnx.html#torch.onnx.export
        to have more information on how to create the ``dict``.

        Returns:
            dict: with the dynamic axes of the input/output.
        """
        return {self._input_name: {0: "batch"}, self._output_name: {0: "batch"}}

    def get_output_names(self) -> tuple[str]:
        r"""Gets the tuple of output names for the network.

        The order of the name should be the same that the order in the
        outputs of the forward function.

        Returns:
            tuple: The tuple of output names.
        """
        return (self._output_name,)



r"""This module implements a training loop using automatic mixed precision
(AMP)."""

__all__ = ["AMPTrainingLoop"]

import logging
from typing import Any, Optional, Union

import torch
from torch.cuda.amp import GradScaler, autocast
from torch.nn import Module
from torch.optim import Optimizer

from gravitorch import constants as ct
from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.utils.device_placement import BaseDevicePlacement
from gravitorch.utils.loop_observers import BaseLoopObserver
from gravitorch.utils.profilers import BaseProfiler
from gravitorch.utils.training_loops.vanilla import VanillaTrainingLoop

logger = logging.getLogger(__name__)


class AMPTrainingLoop(VanillaTrainingLoop):
    r"""Implements a training loop to train a model on a dataset by using
    training loop using automatic mixed precision (AMP).

    Args:
        set_grad_to_none (bool, optional): If ``True``, set the
            gradients to ``None``, otherwise set the gradients to
            zero. Setting the gradients to ``None`` will in general
            have lower memory footprint, and can modestly improve
            performance. Default: ``False``
        amp_enabled (bool, optional): If ``True``, automatic mixed
            precision (AMP) is enabled, otherwise it is disabled.
            Default: ``True``
        batch_device_placement (bool, optional): Specifies the batch
            device placement module. This module moves the batch on
            a target device. The target device should be compatible
            with the model. If ``None``, an ``AutoDevicePlacement``
            object is instantiated. Default: ``None``
        tag (str, optional): Specifies the tag which is used to log
            metrics. Default: ``"train"``
        clip_grad (dict or None, optional): Specifies the
            configuration to clip the gradient. If ``None``, no
            gradient clipping is used during the training.
            Default: ``None``
        observer (``BaseLoopObserver`` or dict or None, optional):
            Specifies the loop observer or its configuration.
            If ``None``, the ``NoOpLoopObserver`` is instantiated.
            Default: ``None``
        profiler (``BaseProfiler`` or dict or None, optional): Specifies
            the profiler or its configuration. If ``None``, the
            ``NoOpProfiler`` is instantiated. Default: ``None``
    """

    def __init__(
        self,
        clip_grad: Optional[dict] = None,
        set_grad_to_none: bool = False,
        amp_enabled: bool = True,
        batch_device_placement: Union[BaseDevicePlacement, dict, None] = None,
        tag: str = "train",
        observer: Union[BaseLoopObserver, dict, None] = None,
        profiler: Union[BaseProfiler, dict, None] = None,
    ):
        super().__init__(
            clip_grad=clip_grad,
            set_grad_to_none=set_grad_to_none,
            batch_device_placement=batch_device_placement,
            tag=tag,
            observer=observer,
            profiler=profiler,
        )
        self._amp_enabled = bool(amp_enabled)
        self._scaler = GradScaler(enabled=self._amp_enabled)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  set_grad_to_none={self._set_grad_to_none},\n"
            f"  amp_enabled={self._amp_enabled},\n"
            f"  tag={self._tag},\n"
            f"  batch_device_placement={self._batch_device_placement},\n"
            f"  clip_grad_fn={self._clip_grad_fn},\n"
            f"  clip_grad_args={self._clip_grad_args},\n"
            f"  observer={self._observer},\n"
            f"  profiler={self._profiler},\n"
            ")"
        )

    def load_state_dict(self, state_dict: dict[str, Any]) -> None:
        self._scaler.load_state_dict(state_dict[ct.SCALER])

    def state_dict(self) -> dict[str, Any]:
        return {ct.SCALER: self._scaler.state_dict()}

    def _train_one_batch(
        self, engine: BaseEngine, model: Module, optimizer: Optimizer, batch: Any
    ) -> dict:
        engine.fire_event(EngineEvents.TRAIN_ITERATION_STARTED)
        optimizer.zero_grad(self._set_grad_to_none)
        with autocast(enabled=self._amp_enabled):
            output = model(self._batch_device_placement.send(batch))
        engine.fire_event(EngineEvents.TRAIN_FORWARD_COMPLETED)

        loss = self._scaler.scale(output[ct.LOSS])
        if torch.isnan(loss):
            logger.warning(
                "NaN detected. The gradient is not computed for this batch "
                f"(iteration: {engine.iteration})"
            )
            engine.fire_event(EngineEvents.TRAIN_ITERATION_COMPLETED)
            return output

        loss.backward()
        if self._clip_grad_fn:
            self._scaler.unscale_(optimizer)
            self._clip_grad_fn(model.parameters(), *self._clip_grad_args)
        engine.fire_event(EngineEvents.TRAIN_BACKWARD_COMPLETED)

        self._scaler.step(optimizer)
        self._scaler.update()
        engine.fire_event(EngineEvents.TRAIN_ITERATION_COMPLETED)

        return output



r"""This package contains the implementation of a lot of utilities used in the
other packages."""

__all__ = ["AssetManager", "get_available_devices", "manual_seed", "move_to_device", "to_tuple"]


from gravitorch.utils.asset_manager import AssetManager
from gravitorch.utils.seed import manual_seed
from gravitorch.utils.sequence import to_tuple
from gravitorch.utils.torch_device import get_available_devices, move_to_device



__all__ = ["MNISTDataset"]

from unittest.mock import Mock

from gravitorch import constants as ct
from gravitorch.utils.integrations import check_torchvision, is_torchvision_available

if is_torchvision_available():
    from torchvision.datasets import MNIST
else:
    MNIST = Mock  # pragma: no cover


class MNISTDataset(MNIST):
    r"""Updated MNIST dataset to return a dict instead of a tuple."""

    def __init__(self, *args, **kwargs):
        check_torchvision()
        super().__init__(*args, **kwargs)

    def __getitem__(self, index: int) -> dict:
        r"""Get the image and the target of the index-th example.

        Args:
            index (int): Specifies the index of the example.

        Returns:
            dict: A dictionary with the image and the target of the
                ``index``-th example.
        """
        img, target = super().__getitem__(index)
        return {ct.INPUT: img, ct.TARGET: target}



__all__ = ["setup_collator"]

from collections.abc import Callable
from typing import Union

from torch.utils.data.dataloader import default_collate

from gravitorch.data.dataloaders.collators.base import BaseCollator


def setup_collator(collator: Union[Callable, dict, None]) -> Callable:
    r"""Sets up a data loader collator.

    Args:
        collator (``Callable`` or dict or None): Specifies the
            data loader collator or its configuration. If ``None``,
            the default data loader collator is used.

    Returns:
        ``Callable``: The data loader collator.
    """
    if collator is None:
        collator = default_collate
    if isinstance(collator, dict):
        collator = BaseCollator.factory(**collator)
    return collator



__all__ = ["is_loss_decreasing", "is_loss_decreasing_with_adam", "is_loss_decreasing_with_sgd"]

import logging
from typing import Any

from torch.nn.utils import clip_grad_norm_
from torch.optim import SGD, Adam, Optimizer

from gravitorch import constants as ct
from gravitorch.models.base import BaseModel
from gravitorch.nn.utils.module_helpers import module_mode
from gravitorch.utils.seed import torch_seed

logger = logging.getLogger(__name__)


def is_loss_decreasing(
    model: BaseModel,
    optimizer: Optimizer,
    batch: Any,
    num_iterations: int = 1,
    random_seed: int = 11409583194270223596,
) -> bool:
    r"""Checks if the loss decreased after some iterations.

    Args:
        model (``BaseModel``): Specifies the model to test.
        optimizer (``torch.optim.Optimizer``): Specifies the optimizer
            to update the weights of the model.
        batch: Specifies the batch used to train the model.
        num_iterations (int, optional): Specifies the number of
            optimization steps. Default: ``1``
        random_seed (int, optional): Specifies the random seed. The
            random seed is used to have a deterministic when the loss
            value is computed. Default: ``11409583194270223596``

    Returns:
        bool: ``True`` if the loss decreased after some iterations,
            otherwise ``False``.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from torch import nn
        >>> from torch.optim import SGD
        >>> from gravitorch.models import VanillaModel
        >>> from gravitorch.models.criterions import VanillaLoss
        >>> from gravitorch.models.networks import BetaMLP
        >>> from gravitorch.models.utils import is_loss_decreasing
        >>> model = VanillaModel(
        ...     network=BetaMLP(input_size=6, hidden_sizes=(8, 4)),
        ...     criterion=VanillaLoss(criterion=nn.MSELoss()),
        ... )
        >>> optimizer = SGD(model.parameters(), lr=0.01)
        >>> is_loss_decreasing(
        ...     model=model,
        ...     optimizer=optimizer,
        ...     batch={'input': torch.randn(2, 6), 'target': torch.randn(2, 4)},
        ... )
        True
    """
    with module_mode(model):
        model.eval()
        with torch_seed(random_seed):
            initial_loss = model(batch)[ct.LOSS].item()
        logger.debug(f"Initial loss value: {initial_loss}")

        model.train()
        for _ in range(num_iterations):
            optimizer.zero_grad()
            output = model(batch)
            output[ct.LOSS].backward()
            clip_grad_norm_(model.parameters(), max_norm=1, norm_type=2)
            optimizer.step()

        model.eval()
        with torch_seed(random_seed):
            final_loss = model(batch)[ct.LOSS].item()
        logger.debug(f"Final loss value: {final_loss}")
    return final_loss < initial_loss


def is_loss_decreasing_with_adam(
    model: BaseModel,
    batch: Any,
    lr: float = 0.0003,
    num_iterations: int = 1,
    random_seed: int = 11409583194270223596,
) -> bool:
    r"""Checks if the loss decreased after some iterations.

    The weights of the model are updated with Adam.

    Args:
        model (``BaseModel``): Specifies the model to test.
        batch: Specifies the batch used to train the model.
        lr (float, optional): Specifies the learning rate.
            Default: ``0.0003``
        num_iterations (int, optional): Specifies the number of
            optimization steps. Default: ``1``
        random_seed (int, optional): Specifies the random seed. The
            random seed is used to have a deterministic when the loss
            value is computed. Default: ``11409583194270223596``

    Returns:
        bool: ``True`` if the loss decreased after some iterations,
            otherwise ``False``.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from torch import nn
        >>> from gravitorch.models import VanillaModel
        >>> from gravitorch.models.criterions import VanillaLoss
        >>> from gravitorch.models.networks import BetaMLP
        >>> from gravitorch.models.utils import is_loss_decreasing_with_adam
        >>> is_loss_decreasing_with_adam(
        ...     model=VanillaModel(
        ...         network=BetaMLP(input_size=6, hidden_sizes=(8, 4)),
        ...         criterion=VanillaLoss(criterion=nn.MSELoss()),
        ...     ),
        ...     lr=0.0003,
        ...     batch={'input': torch.randn(2, 6), 'target': torch.randn(2, 4)},
        ... )
        True
    """
    return is_loss_decreasing(
        model=model,
        optimizer=Adam(model.parameters(), lr=lr),
        batch=batch,
        num_iterations=num_iterations,
        random_seed=random_seed,
    )


def is_loss_decreasing_with_sgd(
    model: BaseModel,
    batch: Any,
    lr: float = 0.01,
    num_iterations: int = 1,
    random_seed: int = 11409583194270223596,
) -> bool:
    r"""Checks if the loss decreased after some iterations.

    The weights of the model are updated with SGD.

    Args:
        model (``BaseModel``): Specifies the model to test.
        batch: Specifies the batch used to train the model.
        lr (float, optional): Specifies the learning rate. Default: ``0.01``
        num_iterations (int, optional): Specifies the number of
            optimization steps. Default: ``1``
        random_seed (int, optional): Specifies the random seed. The
            random seed is used to have a deterministic when the loss
            value is computed. Default: ``11409583194270223596``

    Returns:
        bool: ``True`` if the loss decreased after some iterations,
            otherwise ``False``.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from torch import nn
        >>> from gravitorch.models import VanillaModel
        >>> from gravitorch.models.criterions import VanillaLoss
        >>> from gravitorch.models.networks import BetaMLP
        >>> from gravitorch.models.utils import is_loss_decreasing_with_sgd
        >>> is_loss_decreasing_with_sgd(
        ...     model=VanillaModel(
        ...         network=BetaMLP(input_size=6, hidden_sizes=(8, 4)),
        ...         criterion=VanillaLoss(criterion=nn.MSELoss()),
        ...     ),
        ...     lr=0.01,
        ...     batch={'input': torch.randn(2, 6), 'target': torch.randn(2, 4)},
        ... )
        True
    """
    return is_loss_decreasing(
        model=model,
        optimizer=SGD(model.parameters(), lr=lr),
        batch=batch,
        num_iterations=num_iterations,
        random_seed=random_seed,
    )



__all__ = ["setup_distribution"]

import logging
from typing import Union

from objectory import factory
from torch.distributions import Distribution

from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def setup_distribution(distribution: Union[Distribution, dict]) -> Distribution:
    r"""Sets up a distribution.

    Args:
        distribution (``torch.distributions.Distribution`` or dict
            or None): Specifies the distribution or its configuration.

    Returns:
        ``torch.distributions.Distribution``: The distribution.
    """
    if isinstance(distribution, dict):
        logger.info(
            "Initializing a distribution from its configuration... "
            f"{str_target_object(distribution)}"
        )
        distribution = factory(**distribution)
    return distribution



r"""This module contains some layers to shift and scale a sequence."""

__all__ = ["ShiftScale", "SequenceShiftScale", "shift_scale", "sequence_shift_scale"]

from torch import Tensor
from torch.nn import Module


class ShiftScale(Module):
    r"""Implements a layer to shift and scale an input sequence from a source
    range to a destination range."""

    def forward(self, tensor: Tensor, src_range: Tensor, dst_range: Tensor) -> Tensor:
        r"""Shifts and scales the input sequence from a source range to a
        destination range.

        Args:
            tensor (``torch.Tensor`` of type float and shape
                ``(batch_size, feature_size)``): Specifies the input
                sequence.
            src_range (``torch.Tensor`` of type float and shape
                ``(batch_size, 2, feature_size)``): Specifies the min
                and max values of the source space. ``[:, 0, :]``
                indicates the min values and ``[:, 1, :]`` indicates
                the max values. For a given dimension, the min and
                max values should be different.
            dst_range (``torch.Tensor`` of type float and shape
            ``(batch_size, 2, feature_size)``): Specifies the min
                and max values of the destination space. ``[:, 0, :]``
                indicates the min values and ``[:, 1, :]`` indicates
                the max values. For a given dimension, the min and max
                values should be different.

        Returns:
            ``torch.Tensor`` of type float and shape
                ``(batch_size, feature_size)``: The shifted and scaled
                sequence.
        """
        return shift_scale(tensor, src_range, dst_range)


class SequenceShiftScale(Module):
    r"""Implements a layer to shift and scale sequences from a source range to a
    destination range.

    Args:
        batch_first (bool, optional): Indicates if the first
            dimension is the batch or the sequence. If ``True``, the
            input sequence should have the shape
            ``(batch_size, sequence_length, feature_size)``, otherwise
            ``(sequence_length, batch_size, feature_size)``.
            Default: ``False``
    """

    def __init__(self, batch_first: bool = False):
        super().__init__()
        self._batch_first = bool(batch_first)

    @property
    def batch_first(self) -> bool:
        r"""bool: Indicates if the first dimension is the batch or
        the sequence. If ``True``, the input representation
        should have the shape
        ``(batch_size, sequence_length, feature_size)``, otherwise
        ``(sequence_length, batch_size, feature_size)``.
        """
        return self._batch_first

    def extra_repr(self) -> str:
        return f"batch_first={self._batch_first}"

    def forward(self, sequence: Tensor, src_range: Tensor, dst_range: Tensor) -> Tensor:
        r"""Shifts and scales a sequence from a source range to a destination
        range.

        Args:
            sequence (``torch.Tensor`` of type float and shape
                ``(batch_size, sequence_length, feature_size)`` if
                ``batch_size=False`` or
                ``(batch_size, sequence_length, feature_size)``
                otherwise): Specifies the sequence to shift and scale.
            src_range (``torch.Tensor`` of type float and shape
                ``(batch_size, 2, feature_size)``): Specifies the min
                and max values of the source space. ``[:, 0, :]``
                indicates the min values and ``[:, 1, :]`` indicates
                the max values. For a given dimension, the min and max
                values should be different.
            dst_range (``torch.Tensor`` of type float and shape
                ``(batch_size, 2, feature_size)``): Specifies the min
                and max values of the destination space. ``[:, 0, :]``
                indicates the min values and ``[:, 1, :]`` indicates
                the max values. For a given dimension, the min and max
                values should be different.

        Returns:
            ``torch.Tensor`` of type float and shape
                ``(sequence_length, batch_size, feature_size)`` if
                ``batch_size=False`` or
                ``(batch_size, sequence_length, feature_size)``
                otherwise: The shifted and scaled sequence.
        """
        return sequence_shift_scale(sequence, src_range, dst_range, self._batch_first)


def shift_scale(tensor: Tensor, src_range: Tensor, dst_range: Tensor) -> Tensor:
    r"""Shifts and scales a tensor from a source range to a destination range.

    Args:
        tensor (``torch.Tensor`` of type float and shape
            ``(batch_size, feature_size)``): Specifies the tensor to
            shift and scale.
        src_range (``torch.Tensor`` of type float and shape
            ``(batch_size, 2, feature_size)``): Specifies the min and
            max values of the source space. ``[:, 0, :]`` indicate
            the min values and ``[:, 1, :]`` indicates the max values.
            For a given dimension, the min and max values should be
            different.
        dst_range (``torch.Tensor`` of type float and shape
        ``(batch_size, 2, feature_size)``): Specifies the min and
            max values of the destination space. ``[:, 0, :]``
            indicates the min values and ``[:, 1, :]`` indicates the
            max values. For a given dimension, the min and max values
            should be different.

    Returns:
        ``torch.Tensor`` of type float and shape
            ``(batch_size, feature_size)``: The shifted and scaled
            tensor.
    """
    src_diff = src_range[:, 1] - src_range[:, 0]
    dst_diff = dst_range[:, 1] - dst_range[:, 0]
    scale = dst_diff / src_diff
    return (tensor - src_range[:, 0]) * scale + dst_range[:, 0]


def sequence_shift_scale(
    sequence: Tensor,
    src_range: Tensor,
    dst_range: Tensor,
    batch_first: bool = False,
) -> Tensor:
    r"""Shifts and scales a sequences from a source range to a destination
    range.

    Args:
        sequence (``torch.Tensor`` of type float and shape
            ``(batch_size, sequence_length, feature_size)`` if
            ``batch_size=False`` or
            ``(batch_size, sequence_length, feature_size)``
            otherwise): Specifies the sequence to shift and scale.
        src_range (``torch.Tensor`` of type float and shape
            ``(batch_size, 2, feature_size)``): Specifies the min and
            max values of the source space. ``[:, 0, :]`` indicates
            the min values and ``[:, 1, :]`` indicates the max values.
            For a given dimension, the min and max values should be
            different.
        dst_range (``torch.Tensor`` of type float and shape
            ``(batch_size, 2, feature_size)``): Specifies the min and
            max values of the destination space. ``[:, 0, :]``
            indicates the min values and ``[:, 1, :]`` indicates the
            max values. For a given dimension, the min and max values
            should be different.
        batch_first (bool, optional): Indicates if the first dimension
            is the batch or the sequence. If ``True``, the input
            sequence should have the shape
            ``(batch_size, sequence_length, feature_size)``, otherwise
            ``(sequence_length, batch_size, feature_size)``.
            Default: ``False``

    Returns:
        ``torch.Tensor`` of type float and shape
            ``(sequence_length, batch_size, feature_size)``
            if ``batch_size=False`` or
            ``(batch_size, sequence_length, feature_size)`` otherwise:
            The shifted and scaled tensor.
    """
    src_diff = src_range[:, 1] - src_range[:, 0]
    dst_diff = dst_range[:, 1] - dst_range[:, 0]
    scale = dst_diff / src_diff
    if batch_first:
        return (sequence - src_range[:, None, 0]) * scale[:, None, :] + dst_range[:, None, 0]
    return (sequence - src_range[:, 0]) * scale + dst_range[:, 0]



r"""This module implements some utility functions to automatically do some
operations.

These implementations cover a large set of use cases but you may need to
implement your own functions for some special use cases.
"""

__all__ = ["auto_ddp_model"]

import logging

from torch.nn import Module, SyncBatchNorm
from torch.nn.parallel import DistributedDataParallel

from gravitorch.distributed import comm as dist
from gravitorch.nn.utils.module_helpers import (
    has_learnable_parameters,
    is_module_on_device,
)

logger = logging.getLogger(__name__)


def auto_ddp_model(model: Module, sync_batch_norm: bool = False, **kwargs) -> Module:
    r"""Wraps the model to be ready for Distributed Data Parallel (DDP)
    training.

    Internally, this function does the following steps:

    - Send model to current device if model's parameters are not on
        the device.
    - If NCCL backend, the ``BatchNorm*D`` modules can be converted
        to ``SyncBatchNorm`` modules.
    - Wrap the model to ``DistributedDataParallel`` for native torch
        distributed (NCCL or GLOO) if world size is larger than 1.

    Args:
        model (``torch.nn.Module``): Specifies the model to wrap
            with ``DistributedDataParallel``.
        sync_batch_norm (bool): Specifies if the ``BatchNorm*D``
            modules should be converted to ``SyncBatchNorm`` modules
            (only for NCCL backend).
        **kwargs: Variable arguments for the
            ``DistributedDataParallel`` module. If you use the NCCL
            backend, you cannot set the ``device_ids`` because this
            input is automatically set in this function.

    Returns:
        ``torch.nn.Module``: The model wrapped in a
            ``DistributedDataParallel`` module.
    """
    if isinstance(model, DistributedDataParallel):
        logger.warning(
            "The model is already an instance of ``torch.nn.parallel.DistributedDataParallel`` "
            "so no operation is performed"
        )
        return model
    model = _manage_model_device(model)
    if has_learnable_parameters(model) and dist.get_world_size() > 1:
        model = _wrap_distributed_data_parallel(model, sync_batch_norm, **kwargs)
    return model


def _manage_model_device(model: Module) -> Module:
    r"""Manages the devices of the model.

    Internally, this function moves model's parameters to device
        if its parameters are not on the device.

    Args:
        model (``torch.nn.Module``): Specifies the model to manage
            the devices.
    """
    device = dist.device()
    logger.info(f"Move model parameters to {device}")
    if not is_module_on_device(module=model, device=device):
        model.to(device)
    return model


def _wrap_distributed_data_parallel(
    model: Module, sync_batch_norm: bool = False, **kwargs
) -> Module:
    r"""Wraps the model with the ``DistributedDataParallel`` module.

    Args:
        model (``torch.nn.Module``): Specifies the model to wrap with
            ``DistributedDataParallel``.
        sync_batch_norm (bool): Specifies if the ``BatchNorm*D``
            modules should be converted to ``SyncBatchNorm`` modules
            (only for NCCL backend).
        **kwargs: Variable arguments for the
            ``DistributedDataParallel`` module. You cannot set the
            ``device_ids`` because this input is automatically set
            in this function.

    Returns:
        ``torch.nn.Module``: The model wrapped in a
            ``DistributedDataParallel`` module.
    """
    backend = dist.backend()
    if backend == dist.Backend.NCCL:
        if sync_batch_norm:
            logger.info("Convert batch norm to sync batch norm")
            model = SyncBatchNorm.convert_sync_batchnorm(model)

        lrank = dist.get_local_rank()
        logger.info(
            f"Apply torch DistributedDataParallel on model, device id: {lrank}",
        )
        return DistributedDataParallel(model, device_ids=[lrank], **kwargs)
    if backend == dist.Backend.GLOO:
        logger.info("Apply torch DistributedDataParallel on model")
        return DistributedDataParallel(model, **kwargs)
    return model



r"""This module defines some functionalities to instantiate dynamically a
``torch.utils.data.Dataset`` object from its configuration."""

__all__ = ["setup_dataset"]

import logging
from typing import Optional, TypeVar, Union

from objectory import factory
from torch.utils.data import Dataset

from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)

T = TypeVar("T")


def setup_dataset(dataset: Union[Dataset, dict, None]) -> Optional[Dataset]:
    r"""Sets up a dataset.

    Args:
        dataset (``Dataset`` or dict or ``None``): Specifies the
            dataset or its configuration (dictionary). If a
            configuration is given, a dataset object is instantiated
            from the configuration.

    Returns:
        ``torch.utils.data.Dataset`` or ``None``: A dataset object or
            ``None`` if there is no dataset.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.data.datasets import setup_dataset
        >>> mnist = setup_dataset(
        ...     {'_target_': 'torchvision.datasets.MNIST', 'root': '/my/path/', 'download': True},
        ... )
        >>> mnist
        Dataset MNIST
            Number of datapoints: 60000
            Root location: tmp
            Split: Train
        >>> setup_dataset(mnist)  # Do nothing because the dataset is already instantiated
        Dataset MNIST
            Number of datapoints: 60000
            Root location: tmp
            Split: Train
        >>> setup_dataset(None)
    """
    if isinstance(dataset, dict):
        logger.info(
            f"Initializing a dataset from its configuration... {str_target_object(dataset)}"
        )
        dataset = factory(**dataset)
    return dataset
